import asyncio
from collections.abc import AsyncGenerator, AsyncIterator
from pathlib import PureWindowsPath
from typing import Generic, TypeVar, get_args

from bluesky.protocols import Hints, StreamAsset
from event_model import (  # type: ignore
    ComposeStreamResource,
    DataKey,
    StreamRange,
)
from pydantic import PositiveInt

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    DatasetDescriber,
    DetectorWriter,
    EnableDisable,
    PathInfo,
    PathProvider,
    error_if_none,
    observe_value,
    set_and_wait_for_value,
)
from ophyd_async.epics.core import stop_busy_record

# from ophyd_async.epics.adcore._core_logic import ADBaseDatasetDescriber
from ._core_io import (
    ADBaseDatasetDescriber,
    NDArrayBaseIO,
    NDFileIO,
    NDFilePluginIO,
    NDPluginBaseIO,
)
from ._utils import ADFileWriteMode

NDFileIOT = TypeVar("NDFileIOT", bound=NDFileIO)
ADWriterT = TypeVar("ADWriterT", bound="ADWriter")


class ADWriter(DetectorWriter, Generic[NDFileIOT]):
    """Common behavior for all areaDetector writers."""

    default_suffix: str = "FILE1:"

    def __init__(
        self,
        fileio: NDFileIOT,
        path_provider: PathProvider,
        dataset_describer: DatasetDescriber,
        file_extension: str = "",
        mimetype: str = "",
        plugins: dict[str, NDPluginBaseIO] | None = None,
    ) -> None:
        self._plugins = plugins or {}
        self.fileio = fileio
        self._path_provider: PathProvider = path_provider
        self._path_info: PathInfo | None = None
        self._dataset_describer = dataset_describer
        self._file_extension = file_extension
        self._mimetype = mimetype
        self._last_emitted = 0
        self._emitted_resource = None

        self._capture_status: AsyncStatus | None = None
        self._filename_template = "%s%s_%6.6d"

    @classmethod
    def with_io(
        cls: type[ADWriterT],
        prefix: str,
        path_provider: PathProvider,
        dataset_source: NDArrayBaseIO | None = None,
        fileio_suffix: str | None = None,
        plugins: dict[str, NDPluginBaseIO] | None = None,
    ) -> ADWriterT:
        try:
            fileio_cls = get_args(cls.__orig_bases__[0])[0]  # type: ignore
        except IndexError as err:
            raise RuntimeError("File IO class for writer not specified!") from err

        fileio = fileio_cls(prefix + (fileio_suffix or cls.default_suffix))
        dataset_describer = ADBaseDatasetDescriber(dataset_source or fileio)

        writer = cls(fileio, path_provider, dataset_describer, plugins=plugins)
        return writer

    async def _begin_capture(self, name: str) -> None:
        path_info = error_if_none(
            self._path_info, "Writer must be opened before beginning capture!"
        )

        if isinstance(self.fileio, NDFilePluginIO):
            await self.fileio.enable_callbacks.set(EnableDisable.ENABLE)

        # Set the directory creation depth first, since dir creation callback happens
        # when directory path PV is processed.
        await self.fileio.create_directory.set(path_info.create_dir_depth)

        # Need to ensure that trailing separator is added to the directory path.
        # When setting the path for windows based AD IOCs, a '/' is added rather than
        # a '\\', which will cause the readback to never register the same value.
        dir_path_as_str = str(path_info.directory_path)
        separator = "/"
        if isinstance(path_info.directory_path, PureWindowsPath):
            separator = "\\"

        dir_path_as_str += separator

        await asyncio.gather(
            # See https://github.com/bluesky/ophyd-async/issues/122
            self.fileio.file_path.set(dir_path_as_str),
            self.fileio.file_name.set(path_info.filename),
            self.fileio.file_write_mode.set(ADFileWriteMode.STREAM),
            # For non-HDF file writers, use AD file templating mechanism
            # for generating multi-image datasets
            self.fileio.file_template.set(
                self._filename_template + self._file_extension
            ),
            self.fileio.auto_increment.set(True),
            self.fileio.file_number.set(0),
        )

        if not await self.fileio.file_path_exists.get_value():
            msg = f"Path {dir_path_as_str} doesn't exist or not writable!"
            raise FileNotFoundError(msg)

        # Overwrite num_capture to go forever
        await self.fileio.num_capture.set(0)
        # Wait for it to start, stashing the status that tells us when it finishes
        self._capture_status = await set_and_wait_for_value(
            self.fileio.capture, True, wait_for_set_completion=False
        )

    async def open(
        self, name: str, exposures_per_event: PositiveInt = 1
    ) -> dict[str, DataKey]:
        self._emitted_resource = None
        self._last_emitted = 0
        self._exposures_per_event = exposures_per_event
        frame_shape = await self._dataset_describer.shape()
        dtype_numpy = await self._dataset_describer.np_datatype()

        self._path_info = self._path_provider(device_name=name)

        await self._begin_capture(name)

        describe = {
            name: DataKey(
                source=self.fileio.full_file_name.source,
                shape=[exposures_per_event, *frame_shape],
                dtype="array",
                dtype_numpy=dtype_numpy,
                external="STREAM:",
            )  # type: ignore
        }
        return describe

    async def observe_indices_written(
        self, timeout: float
    ) -> AsyncGenerator[int, None]:
        """Wait until a specific index is ready to be collected."""
        async for num_captured in observe_value(self.fileio.num_captured, timeout):
            yield num_captured // self._exposures_per_event

    async def get_indices_written(self) -> int:
        num_captured = await self.fileio.num_captured.get_value()
        return num_captured // self._exposures_per_event

    async def collect_stream_docs(
        self, name: str, indices_written: int
    ) -> AsyncIterator[StreamAsset]:
        path_info = error_if_none(
            self._path_info, "Writer must be opened before collecting stream docs!"
        )

        if indices_written:
            if not self._emitted_resource:
                file_name = await self.fileio.file_name.get_value()
                file_template = file_name + "_{:06d}" + self._file_extension

                frame_shape = await self._dataset_describer.shape()

                bundler_composer = ComposeStreamResource()

                self._emitted_resource = bundler_composer(
                    mimetype=self._mimetype,
                    uri=str(path_info.directory_uri),
                    # TODO no reference to detector's name
                    data_key=name,
                    parameters={
                        # Assume that we always write 1 frame per file/chunk, this
                        # may change to self._exposures_per_event in the future
                        "chunk_shape": (1, *frame_shape),
                        # Include file template for reconstruction in consolidator
                        "template": file_template,
                    },
                    uid=None,
                    validate=True,
                )

                yield "stream_resource", self._emitted_resource.stream_resource_doc

            # Indices are relative to resource
            if indices_written > self._last_emitted:
                indices: StreamRange = {
                    "start": self._last_emitted,
                    "stop": indices_written,
                }
                self._last_emitted = indices_written
                yield (
                    "stream_datum",
                    self._emitted_resource.compose_stream_datum(indices),
                )

    async def close(self):
        # Already done a caput callback in _capture_status, so can't do one here
        await stop_busy_record(self.fileio.capture, False, timeout=DEFAULT_TIMEOUT)
        if self._capture_status and not self._capture_status.done:
            # We kicked off an open, so wait for it to return
            await self._capture_status
        self._capture_status = None

    def get_hints(self, name: str) -> Hints:
        return {"fields": [name]}



import asyncio
from collections.abc import AsyncGenerator, AsyncIterator

from bluesky.protocols import StreamAsset
from event_model import DataKey  # type: ignore

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    DetectorWriter,
    Device,
    DeviceVector,
    PathProvider,
    Reference,
    SignalR,
    StrictEnum,
    observe_value,
    set_and_wait_for_value,
    wait_for_value,
)
from ophyd_async.epics.core import (
    epics_signal_r,
    epics_signal_rw,
    epics_signal_rw_rbv,
    stop_busy_record,
)


class Writing(StrictEnum):
    CAPTURE = "Capture"
    DONE = "Done"


class OdinNode(Device):
    def __init__(self, prefix: str, name: str = "") -> None:
        self.writing = epics_signal_r(str, f"{prefix}Writing_RBV")
        self.frames_dropped = epics_signal_r(int, f"{prefix}FramesDropped_RBV")
        self.frames_time_out = epics_signal_r(int, f"{prefix}FramesTimedOut_RBV")
        self.error_status = epics_signal_r(str, f"{prefix}FPErrorState_RBV")
        self.fp_initialised = epics_signal_r(int, f"{prefix}FPProcessConnected_RBV")
        self.fr_initialised = epics_signal_r(int, f"{prefix}FRProcessConnected_RBV")
        self.num_captured = epics_signal_r(int, f"{prefix}NumCaptured_RBV")
        self.clear_errors = epics_signal_rw(int, f"{prefix}FPClearErrors")
        self.error_message = epics_signal_rw(str, f"{prefix}FPErrorMessage_RBV")

        super().__init__(name)


class Odin(Device):
    def __init__(self, prefix: str, name: str = "", nodes: int = 4) -> None:
        # default nodes is set to 4, MX 16M Eiger detectors - nodes = 4.
        # B21 4M Eiger detector - nodes = 1
        self.nodes = DeviceVector(
            {i: OdinNode(f"{prefix[:-1]}{i + 1}:") for i in range(nodes)}
        )

        self.capture = epics_signal_rw(Writing, f"{prefix}Capture")
        self.capture_rbv = epics_signal_r(str, prefix + "Capture_RBV")
        self.num_captured = epics_signal_r(int, f"{prefix}NumCaptured_RBV")
        self.num_to_capture = epics_signal_rw_rbv(int, f"{prefix}NumCapture")

        self.start_timeout = epics_signal_rw(str, f"{prefix}StartTimeout")
        self.timeout_active_rbv = epics_signal_r(str, f"{prefix}TimeoutActive_RBV")

        self.image_height = epics_signal_rw_rbv(int, f"{prefix}ImageHeight")
        self.image_width = epics_signal_rw_rbv(int, f"{prefix}ImageWidth")

        self.num_row_chunks = epics_signal_rw_rbv(int, f"{prefix}NumRowChunks")
        self.num_col_chunks = epics_signal_rw_rbv(int, f"{prefix}NumColChunks")

        self.file_path = epics_signal_rw_rbv(str, f"{prefix}FilePath")
        self.file_name = epics_signal_rw_rbv(str, f"{prefix}FileName")
        self.id = epics_signal_r(str, f"{prefix}AcquisitionID_RBV")

        self.num_frames_chunks = epics_signal_rw(int, prefix + "NumFramesChunks")
        self.meta_active = epics_signal_r(str, prefix + "META:AcquisitionActive_RBV")
        self.meta_writing = epics_signal_r(str, prefix + "META:Writing_RBV")
        self.meta_file_name = epics_signal_r(str, f"{prefix}META:FileName_RBV")
        self.meta_stop = epics_signal_rw(bool, f"{prefix}META:Stop")

        self.fan_ready = epics_signal_rw(float, f"{prefix}FAN:StateReady_RBV")

        self.data_type = epics_signal_rw_rbv(str, f"{prefix}DataType")

        super().__init__(name)


class OdinWriter(DetectorWriter):
    def __init__(
        self,
        path_provider: PathProvider,
        odin_driver: Odin,
        detector_bit_depth: SignalR[int],
    ) -> None:
        self._drv = odin_driver
        self._path_provider = path_provider
        self._detector_bit_depth = Reference(detector_bit_depth)
        self._capture_status: AsyncStatus | None = None
        super().__init__()

    async def open(self, name: str, exposures_per_event: int = 1) -> dict[str, DataKey]:
        info = self._path_provider(device_name=name)
        self._exposures_per_event = exposures_per_event

        await asyncio.gather(
            self._drv.data_type.set(
                f"UInt{await self._detector_bit_depth().get_value()}"
            ),
            self._drv.num_to_capture.set(0),
            self._drv.file_path.set(str(info.directory_path)),
            self._drv.file_name.set(info.filename),
        )

        await asyncio.gather(
            wait_for_value(
                self._drv.meta_file_name, info.filename, timeout=DEFAULT_TIMEOUT
            ),
            wait_for_value(self._drv.id, info.filename, timeout=DEFAULT_TIMEOUT),
            wait_for_value(self._drv.meta_active, "Active", timeout=DEFAULT_TIMEOUT),
        )

        self._capture_status = await set_and_wait_for_value(
            self._drv.capture, Writing.CAPTURE, wait_for_set_completion=False
        )

        await asyncio.gather(
            wait_for_value(self._drv.capture_rbv, "Capturing", timeout=DEFAULT_TIMEOUT),
            wait_for_value(self._drv.meta_writing, "Writing", timeout=DEFAULT_TIMEOUT),
        )

        return await self._describe()

    async def _describe(self) -> dict[str, DataKey]:
        data_shape = await asyncio.gather(
            self._drv.image_height.get_value(), self._drv.image_width.get_value()
        )

        return {
            "data": DataKey(
                source=self._drv.file_name.source,
                shape=[self._exposures_per_event, *data_shape],
                dtype="array",
                # TODO: Use correct type based on eiger https://github.com/bluesky/ophyd-async/issues/529
                dtype_numpy="<u2",
                external="STREAM:",
            )
        }

    async def observe_indices_written(
        self, timeout: float
    ) -> AsyncGenerator[int, None]:
        async for num_captured in observe_value(self._drv.num_captured, timeout):
            yield num_captured // self._exposures_per_event

    async def get_indices_written(self) -> int:
        return await self._drv.num_captured.get_value() // self._exposures_per_event

    def collect_stream_docs(
        self, name: str, indices_written: int
    ) -> AsyncIterator[StreamAsset]:
        # TODO: Correctly return stream https://github.com/bluesky/ophyd-async/issues/530
        raise NotImplementedError()

    async def close(self) -> None:
        await stop_busy_record(self._drv.capture, Writing.DONE, timeout=DEFAULT_TIMEOUT)
        await self._drv.meta_stop.set(True, wait=True)
        if self._capture_status and not self._capture_status.done:
            await self._capture_status
        self._capture_status = None



import pytest

from ophyd_async.core import init_devices
from ophyd_async.epics.motor import Motor
from ophyd_async.epics.pmac import PmacIO
from ophyd_async.testing import set_mock_value


@pytest.fixture
async def sim_motors():
    async with init_devices(mock=True):
        sim_x_motor = Motor("BLxxI-MO-STAGE-01:X")
        sim_y_motor = Motor("BLxxI-MO-STAGE-01:Y")
        sim_pmac = PmacIO(
            prefix="Test_PMAC",
            raw_motors=[sim_x_motor, sim_y_motor],
            coord_nums=[1],
        )

    pmac_x = sim_pmac.assignment[sim_pmac.motor_assignment_index[sim_x_motor]]
    pmac_y = sim_pmac.assignment[sim_pmac.motor_assignment_index[sim_y_motor]]
    set_mock_value(pmac_x.cs_port, "CS1")
    set_mock_value(pmac_x.cs_number, 1)
    set_mock_value(pmac_x.cs_axis_letter, "X")
    set_mock_value(pmac_y.cs_port, "CS1")
    set_mock_value(pmac_y.cs_number, 1)
    set_mock_value(pmac_y.cs_axis_letter, "Y")
    set_mock_value(sim_x_motor.acceleration_time, 0.5)
    set_mock_value(sim_x_motor.velocity, 1)
    set_mock_value(sim_x_motor.max_velocity, 5)
    set_mock_value(sim_y_motor.acceleration_time, 0.5)
    set_mock_value(sim_y_motor.velocity, 1)
    set_mock_value(sim_y_motor.max_velocity, 10)
    yield (sim_pmac, sim_x_motor, sim_y_motor)



import warnings
from collections.abc import Awaitable, Callable, Generator, Sequence
from contextlib import contextmanager
from enum import Enum
from typing import Any, cast

from bluesky.protocols import HasHints, Hints, Reading
from event_model import DataKey

from ._device import Device, DeviceVector
from ._protocol import AsyncConfigurable, AsyncReadable, AsyncStageable
from ._signal import SignalR
from ._status import AsyncStatus
from ._utils import merge_gathered_dicts


class StandardReadableFormat(Enum):
    """Declare how a `Device` should contribute to the `StandardReadable` verbs."""

    CHILD = "CHILD"
    """Detect which verbs the child supports and contribute to:

    - `read()`, `describe()` if it is [](#bluesky.protocols.Readable)
    - `read_configuration()`, `describe_configuration()` if it is
      [](#bluesky.protocols.Configurable)
    - `stage()`, `unstage()` if it is [](#bluesky.protocols.Stageable)
    - `hints` if it [](#bluesky.protocols.HasHints)
    """

    CONFIG_SIGNAL = "CONFIG_SIGNAL"
    """Contribute the `Signal` value to `read_configuration()` and
    `describe_configuration()`
    """

    HINTED_SIGNAL = "HINTED_SIGNAL"
    """Contribute the monitored `Signal` value to `read()` and `describe()` and
    put the signal name in `hints`
    """

    UNCACHED_SIGNAL = "UNCACHED_SIGNAL"
    """Contribute the uncached `Signal` value to `read()` and `describe()`
    """

    HINTED_UNCACHED_SIGNAL = "HINTED_UNCACHED_SIGNAL"
    """Contribute the uncached `Signal` value to `read()` and `describe()` and
    put the signal name in `hints`
    """

    def __call__(self, parent: Device, child: Device):
        if not isinstance(parent, StandardReadable):
            raise TypeError(f"Expected parent to be StandardReadable, got {parent}")
        parent.add_readables([child], self)


# Back compat
class _WarningMatcher:
    def __init__(self, name: str, target: StandardReadableFormat):
        self._name = name
        self._target = target

    def __eq__(self, value: object) -> bool:
        warnings.warn(
            DeprecationWarning(
                f"Use `StandardReadableFormat.{self._target.name}` "
                f"instead of `{self._name}`"
            ),
            stacklevel=2,
        )
        return value == self._target


def _compat_format(name: str, target: StandardReadableFormat) -> StandardReadableFormat:
    return cast(StandardReadableFormat, _WarningMatcher(name, target))


ConfigSignal = _compat_format("ConfigSignal", StandardReadableFormat.CONFIG_SIGNAL)
HintedSignal: Any = _compat_format("HintedSignal", StandardReadableFormat.HINTED_SIGNAL)
HintedSignal.uncached = _compat_format(
    "HintedSignal.uncached", StandardReadableFormat.HINTED_UNCACHED_SIGNAL
)


class StandardReadable(
    Device, AsyncReadable, AsyncConfigurable, AsyncStageable, HasHints
):
    """Device that provides selected child Device values in `read()`.

    Provides the ability for children to be registered to:
    - Participate in `stage()` and `unstage()`
    - Provide their value in `read()` and `describe()
    - Provide their value in `read_configuration()` and `describe_configuration()
    - Select a value to appear in `hints`

    The behavior is customized with a [](#StandardReadableFormat)
    """

    # These must be immutable types to avoid accidental sharing between
    # different instances of the class
    _describe_config_funcs: tuple[Callable[[], Awaitable[dict[str, DataKey]]], ...] = ()
    _read_config_funcs: tuple[Callable[[], Awaitable[dict[str, Reading]]], ...] = ()
    _describe_funcs: tuple[Callable[[], Awaitable[dict[str, DataKey]]], ...] = ()
    _read_funcs: tuple[Callable[[], Awaitable[dict[str, Reading]]], ...] = ()
    _stageables: tuple[AsyncStageable, ...] = ()
    _has_hints: tuple[HasHints, ...] = ()

    @AsyncStatus.wrap
    async def stage(self) -> None:
        for sig in self._stageables:
            await sig.stage().task

    @AsyncStatus.wrap
    async def unstage(self) -> None:
        for sig in self._stageables:
            await sig.unstage().task

    async def describe_configuration(self) -> dict[str, DataKey]:
        return await merge_gathered_dicts(
            [func() for func in self._describe_config_funcs]
        )

    async def read_configuration(self) -> dict[str, Reading]:
        return await merge_gathered_dicts([func() for func in self._read_config_funcs])

    async def describe(self) -> dict[str, DataKey]:
        return await merge_gathered_dicts([func() for func in self._describe_funcs])

    async def read(self) -> dict[str, Reading]:
        return await merge_gathered_dicts([func() for func in self._read_funcs])

    @property
    def hints(self) -> Hints:
        hints: Hints = {}
        for new_hint in self._has_hints:
            # Merge the existing and new hints, based on the type of the value.
            # This avoids default dict merge behavior that overrides the values;
            # we want to combine them when they are Sequences, and ensure they are
            # identical when string values.
            for key, value in new_hint.hints.items():
                # fail early for unkwon types
                if isinstance(value, str):
                    if key in hints:
                        if hints[key] != value:
                            msg = f"Hints key {key} value may not be overridden"
                            raise RuntimeError(msg)
                    else:
                        hints[key] = value  # type: ignore[literal-required]
                elif isinstance(value, Sequence):
                    if key in hints:
                        for new_val in value:
                            if new_val in hints[key]:
                                msg = f"Hint {key} {new_val} overrides existing hint"
                                raise RuntimeError(msg)
                        hints[key] = (  # type: ignore[literal-required]
                            hints[key] + value  # type: ignore[literal-required]
                        )
                    else:
                        hints[key] = value  # type: ignore[literal-required]
                else:
                    msg = (
                        f"{new_hint.name}: Unknown type for value '{value}'"
                        f" for key '{key}'"
                    )
                    raise TypeError(msg)

        return hints

    @contextmanager
    def add_children_as_readables(
        self,
        format: StandardReadableFormat = StandardReadableFormat.CHILD,
    ) -> Generator[None, None, None]:
        """Context manager that calls [](#add_readables) on child Devices added within.

        Scans `self.children()` on entry and exit to context manager, and calls
        `add_readables()` on any that are added with the provided
        `StandardReadableFormat`.
        """
        dict_copy = dict(self.children())

        yield

        # Set symmetric difference operator gives all newly added keys.
        new_dict = dict(self.children())
        for key, value in new_dict.items():
            # Check if key already exists in dict_copy and if the value has changed.
            if key in dict_copy and value != dict_copy[key]:
                error_msg = (
                    f"Duplicate readable device found: '{key}' in {value.parent}. "
                    "Derived class must not redefine a readable. "
                    "See: https://github.com/bluesky/ophyd-async/issues/848. "
                    "If this functionality is required, please raise an issue: "
                    "https://github.com/bluesky/ophyd-async"
                )
                raise KeyError(error_msg)

        new_keys = dict_copy.keys() ^ new_dict.keys()
        new_values = [new_dict[key] for key in new_keys]

        flattened_values = []
        for value in new_values:
            if isinstance(value, DeviceVector):
                flattened_values.extend(value.values())
            else:
                flattened_values.append(value)

        new_devices = list(filter(lambda x: isinstance(x, Device), flattened_values))
        self.add_readables(new_devices, format)

    def add_readables(
        self,
        devices: Sequence[Device],
        format: StandardReadableFormat = StandardReadableFormat.CHILD,
    ) -> None:
        """Add devices to contribute to various bluesky verbs.

        Use output from the given devices to contribute to the verbs of the following
        interfaces:

        - [](#bluesky.protocols.Readable)
        - [](#bluesky.protocols.Configurable)
        - [](#bluesky.protocols.Stageable)
        - [](#bluesky.protocols.HasHints)

        :param devices: The devices to be added
        :param format:
            Determines which of the devices functions are added to which verb as
            per the [](#StandardReadableFormat) documentation
        """

        def assert_device_is_signalr(device: Device) -> SignalR:
            if not isinstance(device, SignalR):
                raise TypeError(f"{device} is not a SignalR")
            return device

        for device in devices:
            match format:
                case StandardReadableFormat.CHILD:
                    if isinstance(device, AsyncConfigurable):
                        self._describe_config_funcs += (device.describe_configuration,)
                        self._read_config_funcs += (device.read_configuration,)
                    if isinstance(device, AsyncReadable):
                        self._describe_funcs += (device.describe,)
                        self._read_funcs += (device.read,)
                    if isinstance(device, AsyncStageable):
                        self._stageables += (device,)
                    if isinstance(device, HasHints):
                        self._has_hints += (device,)
                case StandardReadableFormat.CONFIG_SIGNAL:
                    signalr_device = assert_device_is_signalr(device=device)
                    self._describe_config_funcs += (signalr_device.describe,)
                    self._read_config_funcs += (signalr_device.read,)
                case StandardReadableFormat.HINTED_SIGNAL:
                    signalr_device = assert_device_is_signalr(device=device)
                    self._describe_funcs += (signalr_device.describe,)
                    self._read_funcs += (signalr_device.read,)
                    self._stageables += (signalr_device,)
                    self._has_hints += (_HintsFromName(signalr_device),)
                case StandardReadableFormat.UNCACHED_SIGNAL:
                    signalr_device = assert_device_is_signalr(device=device)
                    self._describe_funcs += (signalr_device.describe,)
                    self._read_funcs += (_UncachedRead(signalr_device),)
                case StandardReadableFormat.HINTED_UNCACHED_SIGNAL:
                    signalr_device = assert_device_is_signalr(device=device)
                    self._describe_funcs += (signalr_device.describe,)
                    self._read_funcs += (_UncachedRead(signalr_device),)
                    self._has_hints += (_HintsFromName(signalr_device),)


class _UncachedRead:
    def __init__(self, signal: SignalR) -> None:
        self.signal = signal

    async def __call__(self) -> dict[str, Reading]:
        return await self.signal.read(cached=False)


class _HintsFromName(HasHints):
    def __init__(self, device: Device) -> None:
        self.device = device

    @property
    def name(self) -> str:
        return self.device.name

    @property
    def hints(self) -> Hints:
        fields = [self.name] if self.name else []
        return {"fields": fields}



from bluesky.protocols import Reading

from ophyd_async.core import StandardReadable
from ophyd_async.sim._pattern_generator import PatternGenerator

from ._motor import SimMotor


class SimStage(StandardReadable):
    """A simulated sample stage with X and Y movables."""

    def __init__(self, pattern_generator: PatternGenerator, name="") -> None:
        self.pattern_generator = pattern_generator
        # Define some child Devices
        with self.add_children_as_readables():
            self.x = SimMotor(instant=False)
            self.y = SimMotor(instant=False)
        # Set name of device and child devices
        super().__init__(name=name)

    def _set_x_from_reading(self, readings: dict[str, Reading[float]]):
        (x_reading,) = readings.values()
        self.pattern_generator.set_x(x_reading["value"])

    def _set_y_from_reading(self, readings: dict[str, Reading[float]]):
        (y_reading,) = readings.values()
        self.pattern_generator.set_y(y_reading["value"])

    def stage(self):
        """Stage the motors and report the position to the pattern generator."""
        # Tell the pattern generator about the motor positions
        self.x.user_readback.subscribe_reading(self._set_x_from_reading)
        self.y.user_readback.subscribe_reading(self._set_y_from_reading)
        return super().stage()

    def unstage(self):
        """Unstage the motors and remove the position subscription."""
        self.x.user_readback.clear_sub(self._set_x_from_reading)
        self.y.user_readback.clear_sub(self._set_y_from_reading)
        return super().unstage()



"""Core components of the areaDetector software.

https://github.com/areaDetector/ADCore
"""

from ._core_detector import AreaDetector, ContAcqAreaDetector
from ._core_io import (
    ADBaseDatasetDescriber,
    ADBaseIO,
    ADCompression,
    ADState,
    NDArrayBaseIO,
    NDCBFlushOnSoftTrgMode,
    NDFileHDFIO,
    NDFileIO,
    NDFilePluginIO,
    NDPluginBaseIO,
    NDPluginCBIO,
    NDPluginStatsIO,
    NDROIStatIO,
)
from ._core_logic import DEFAULT_GOOD_STATES, ADBaseContAcqController, ADBaseController
from ._core_writer import ADWriter
from ._hdf_writer import ADHDFWriter
from ._jpeg_writer import ADJPEGWriter
from ._single_trigger import SingleTriggerDetector
from ._tiff_writer import ADTIFFWriter
from ._utils import (
    ADBaseDataType,
    ADFileWriteMode,
    ADImageMode,
    NDAttributeDataType,
    NDAttributeParam,
    NDAttributePv,
    NDAttributePvDbrType,
    ndattributes_to_xml,
)

__all__ = [
    "ADBaseIO",
    "ADCompression",
    "ADBaseContAcqController",
    "AreaDetector",
    "ADState",
    "ContAcqAreaDetector",
    "NDArrayBaseIO",
    "NDFileIO",
    "NDFilePluginIO",
    "NDFileHDFIO",
    "NDPluginBaseIO",
    "NDPluginStatsIO",
    "NDROIStatIO",
    "DEFAULT_GOOD_STATES",
    "ADBaseDatasetDescriber",
    "ADBaseController",
    "ADWriter",
    "ADHDFWriter",
    "ADTIFFWriter",
    "ADJPEGWriter",
    "SingleTriggerDetector",
    "ADBaseDataType",
    "ADFileWriteMode",
    "ADImageMode",
    "NDAttributePv",
    "NDAttributeParam",
    "NDAttributeDataType",
    "NDAttributePvDbrType",
    "NDCBFlushOnSoftTrgMode",
    "NDPluginCBIO",
    "ndattributes_to_xml",
]



from ._base_device import TangoDevice, TangoDeviceConnector, TangoPolling
from ._signal import (
    infer_python_type,
    infer_signal_type,
    make_backend,
    tango_signal_r,
    tango_signal_rw,
    tango_signal_w,
    tango_signal_x,
)
from ._tango_readable import TangoReadable
from ._tango_transport import (
    AttributeProxy,
    CommandProxy,
    CommandProxyReadCharacter,
    TangoDoubleStringTable,
    TangoLongStringTable,
    TangoSignalBackend,
    ensure_proper_executor,
    get_command_character,
    get_dtype_extended,
    get_python_type,
    get_source_metadata,
    get_tango_trl,
)
from ._utils import (
    DevStateEnum,
    get_device_trl_and_attr,
    get_full_attr_trl,
    try_to_cast_as_float,
)

__all__ = [
    "AttributeProxy",
    "CommandProxy",
    "CommandProxyReadCharacter",
    "DevStateEnum",
    "ensure_proper_executor",
    "TangoSignalBackend",
    "get_command_character",
    "get_python_type",
    "get_dtype_extended",
    "get_source_metadata",
    "get_tango_trl",
    "infer_python_type",
    "infer_signal_type",
    "make_backend",
    "tango_signal_r",
    "tango_signal_rw",
    "tango_signal_w",
    "tango_signal_x",
    "TangoDevice",
    "TangoReadable",
    "TangoPolling",
    "TangoDeviceConnector",
    "TangoLongStringTable",
    "TangoDoubleStringTable",
    "try_to_cast_as_float",
    "get_device_trl_and_attr",
    "get_full_attr_trl",
]



from pathlib import Path

import pytest

from ophyd_async.core import (
    AsyncReadable,
    StandardFlyer,
)
from ophyd_async.sim import SimMotor


@pytest.mark.parametrize(
    "sim_detector", [("test", "test det", Path("/tmp"))], indirect=True
)
async def test_readable(sim_detector):
    assert isinstance(SimMotor, AsyncReadable)
    assert isinstance(sim_detector, AsyncReadable)
    assert not isinstance(StandardFlyer, AsyncReadable)



from typing import cast

import pytest

from ophyd_async.core import (
    DetectorTrigger,
    TriggerInfo,
)
from ophyd_async.epics import adkinetix
from ophyd_async.testing import set_mock_value


@pytest.fixture
def test_adkinetix(ad_standard_det_factory) -> adkinetix.KinetixDetector:
    return ad_standard_det_factory(adkinetix.KinetixDetector)


async def test_get_deadtime(
    test_adkinetix: adkinetix.KinetixDetector,
):
    # Currently Kinetix driver doesn't support getting deadtime.
    assert test_adkinetix._controller.get_deadtime(0) == 0.001


async def test_trigger_modes(test_adkinetix: adkinetix.KinetixDetector):
    driver = cast(adkinetix.KinetixDriverIO, test_adkinetix.driver)
    set_mock_value(driver.trigger_mode, adkinetix.KinetixTriggerMode.INTERNAL)

    async def setup_trigger_mode(trig_mode: DetectorTrigger):
        await test_adkinetix._controller.prepare(
            TriggerInfo(number_of_events=1, trigger=trig_mode)
        )
        await test_adkinetix._controller.arm()
        await test_adkinetix._controller.wait_for_idle()
        # Prevent timeouts
        set_mock_value(driver.acquire, True)

    # Default TriggerSource
    assert (await driver.trigger_mode.get_value()) == "Internal"

    await setup_trigger_mode(DetectorTrigger.EDGE_TRIGGER)
    assert (await driver.trigger_mode.get_value()) == "Rising Edge"

    await setup_trigger_mode(DetectorTrigger.CONSTANT_GATE)
    assert (await driver.trigger_mode.get_value()) == "Exp. Gate"

    await setup_trigger_mode(DetectorTrigger.INTERNAL)
    assert (await driver.trigger_mode.get_value()) == "Internal"

    await setup_trigger_mode(DetectorTrigger.VARIABLE_GATE)
    assert (await driver.trigger_mode.get_value()) == "Exp. Gate"



"""Module which defines abstract classes to work with detectors."""

import asyncio
import time
from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, AsyncIterator, Callable, Iterator, Sequence
from enum import Enum
from functools import cached_property
from typing import (
    Generic,
    TypeVar,
)

from bluesky.protocols import (
    Collectable,
    Flyable,
    Hints,
    Preparable,
    Reading,
    Stageable,
    StreamAsset,
    Triggerable,
    WritesStreamAssets,
)
from event_model import DataKey
from pydantic import Field, NonNegativeInt, PositiveInt, computed_field

from ._device import Device, DeviceConnector
from ._protocol import AsyncConfigurable, AsyncReadable
from ._signal import SignalR
from ._status import AsyncStatus, WatchableAsyncStatus
from ._utils import DEFAULT_TIMEOUT, ConfinedModel, WatcherUpdate, merge_gathered_dicts


class DetectorTrigger(Enum):
    """Type of mechanism for triggering a detector to take frames."""

    INTERNAL = "INTERNAL"
    """Detector generates internal trigger for given rate"""

    EDGE_TRIGGER = "EDGE_TRIGGER"
    """Expect a series of arbitrary length trigger signals"""

    CONSTANT_GATE = "CONSTANT_GATE"
    """Expect a series of constant width external gate signals"""

    VARIABLE_GATE = "VARIABLE_GATE"
    """Expect a series of variable width external gate signals"""


class TriggerInfo(ConfinedModel):
    """Minimal set of information required to setup triggering on a detector."""

    number_of_events: NonNegativeInt | list[NonNegativeInt] = Field(default=1)
    """Number of events that will be processed, (0 means infinite).

    Can be:
    - A single integer or
    - A list of integers for multiple events

    Example for tomography: ``TriggerInfo(number_of_events=[2,3,100,3])``.
    This would process:

    - 2 events for dark field images
    - 3 events for initial flat field images
    - 100 events for projections
    - 3 events for final flat field images
    """

    trigger: DetectorTrigger = Field(default=DetectorTrigger.INTERNAL)
    """Sort of triggers that will be sent"""

    deadtime: float = Field(default=0.0, ge=0)
    """What is the minimum deadtime between exposures"""

    livetime: float | None = Field(default=None, ge=0)
    """What is the maximum high time of the exposures"""

    exposure_timeout: float | None = Field(default=None, gt=0)
    """What is the maximum timeout on waiting for an exposure"""

    exposures_per_event: PositiveInt = 1
    """The number of exposures that are grouped into a single StreamDatum index.
    A exposures_per_event > 1 can be useful to have exposures from a faster detector
    able to be zipped with a single exposure from a slower detector. E.g. if
    number_of_events=10 and exposures_per_event=5 then the detector will take
    50 exposures, but publish 10 StreamDatum indices, and describe() will show a
    shape of (5, h, w) for each.
    Default is 1.
    """

    @computed_field
    @cached_property
    def total_number_of_exposures(self) -> int:
        return (
            sum(self.number_of_events)
            if isinstance(self.number_of_events, list)
            else self.number_of_events
        ) * self.exposures_per_event


class DetectorController(ABC):
    """Detector logic for arming and disarming the detector."""

    @abstractmethod
    def get_deadtime(self, exposure: float | None) -> float:
        """Get state-independent deadtime.

        For a given exposure, what is the safest minimum time between exposures that
        can be determined without reading signals.
        """

    @abstractmethod
    async def prepare(self, trigger_info: TriggerInfo) -> None:
        """Do all necessary steps to prepare the detector for triggers.

        :param trigger_info: The sort of triggers to expect.
        """

    @abstractmethod
    async def arm(self) -> None:
        """Arm the detector."""

    @abstractmethod
    async def wait_for_idle(self):
        """Wait on the internal _arm_status and wait for it to get disarmed/idle."""

    @abstractmethod
    async def disarm(self):
        """Disarm the detector, return detector to an idle state."""


class DetectorWriter(ABC):
    """Logic for making detector write data to somewhere persistent (e.g. HDF5 file)."""

    @abstractmethod
    async def open(
        self, name: str, exposures_per_event: PositiveInt = 1
    ) -> dict[str, DataKey]:
        """Open writer and wait for it to be ready for data.

        :param exposures_per_event:
            Each StreamDatum index corresponds to this many written exposures
        :return: Output for ``describe()``
        """

    def get_hints(self, name: str) -> Hints:
        """The hints to be used for the detector."""
        return {}

    @abstractmethod
    async def get_indices_written(self) -> int:
        """Get the number of indices written."""

    # Note: this method is really async, but if we make it async here then we
    # need to give it a body with a redundant yield statement, which is a bit
    # awkward. So we just leave it as a regular method and let the user
    # implement it as async.
    @abstractmethod
    def observe_indices_written(self, timeout: float) -> AsyncGenerator[int, None]:
        """Yield the index of each frame (or equivalent data point) as it is written."""

    @abstractmethod
    def collect_stream_docs(
        self, name: str, indices_written: int
    ) -> AsyncIterator[StreamAsset]:
        """Create Stream docs up to given number written."""

    @abstractmethod
    async def close(self) -> None:
        """Close writer, blocks until I/O is complete."""


# Add type var for controller so we can define
# StandardDetector[KinetixController, ADWriter] for example
DetectorControllerT = TypeVar("DetectorControllerT", bound=DetectorController)
DetectorWriterT = TypeVar("DetectorWriterT", bound=DetectorWriter)


def _ensure_trigger_info_exists(trigger_info: TriggerInfo | None) -> TriggerInfo:
    # make absolute sure we realy have a valid TriggerInfo ... mostly for pylance
    if trigger_info is None:
        raise RuntimeError("Trigger info must be set before calling this method.")
    return trigger_info


class StandardDetector(
    Device,
    Stageable,
    AsyncConfigurable,
    AsyncReadable,
    Triggerable,
    Preparable,
    Flyable,
    Collectable,
    WritesStreamAssets,
    Generic[DetectorControllerT, DetectorWriterT],
):
    """Detector base class for step and fly scanning detectors.

    Aggregates controller and writer logic together.

    :param controller: Logic for arming and disarming the detector
    :param writer: Logic for making the detector write persistent data
    :param config_sigs: Signals to read when describe and read configuration are called
    :param name: Device name
    """

    def __init__(
        self,
        controller: DetectorControllerT,
        writer: DetectorWriterT,
        config_sigs: Sequence[SignalR] = (),
        name: str = "",
        connector: DeviceConnector | None = None,
    ) -> None:
        self._controller = controller
        self._writer = writer
        self._describe: dict[str, DataKey] = {}
        self._config_sigs = list(config_sigs)
        # For prepare
        self._arm_status: AsyncStatus | None = None
        self._trigger_info: TriggerInfo | None = None
        # For kickoff
        self._watchers: list[Callable] = []
        self._fly_status: WatchableAsyncStatus | None = None
        self._fly_start: float | None = None
        self._events_to_complete: int = 0
        # Represents the total number of exposures that will have been completed at the
        # end of the next `complete`.
        self._completable_exposures: int = 0
        self._number_of_events_iter: Iterator[int] | None = None
        self._initial_frame: int = 0
        super().__init__(name, connector=connector)

    @AsyncStatus.wrap
    async def stage(self) -> None:
        """Make sure the detector is idle and ready to be used."""
        await self._check_config_sigs()
        await asyncio.gather(self._writer.close(), self._controller.disarm())
        self._trigger_info = None

    async def _check_config_sigs(self):
        """Check configuration signals are named and connected."""
        for signal in self._config_sigs:
            if signal.name == "":
                raise Exception(
                    "config signal must be named before it is passed to the detector"
                )
            try:
                await signal.get_value()
            except NotImplementedError as exc:
                raise Exception(
                    f"config signal {signal.name} must be connected before it is "
                    + "passed to the detector"
                ) from exc

    @AsyncStatus.wrap
    async def unstage(self) -> None:
        """Disarm the detector and stop file writing."""
        await asyncio.gather(self._writer.close(), self._controller.disarm())

    async def read_configuration(self) -> dict[str, Reading]:
        return await merge_gathered_dicts(sig.read() for sig in self._config_sigs)

    async def describe_configuration(self) -> dict[str, DataKey]:
        return await merge_gathered_dicts(sig.describe() for sig in self._config_sigs)

    async def read(self) -> dict[str, Reading]:
        """There is no data to be placed in events, so this is empty."""
        # All data is in StreamResources, not Events, so nothing to output here
        return {}

    async def describe(self) -> dict[str, DataKey]:
        return self._describe

    @AsyncStatus.wrap
    async def trigger(self) -> None:
        if self._trigger_info is None:
            await self.prepare(
                TriggerInfo(
                    number_of_events=1,
                    trigger=DetectorTrigger.INTERNAL,
                )
            )
        trigger_info = _ensure_trigger_info_exists(self._trigger_info)
        if trigger_info.trigger is not DetectorTrigger.INTERNAL:
            msg = "The trigger method can only be called with INTERNAL triggering"
            raise ValueError(msg)
        if trigger_info.number_of_events != 1:
            raise ValueError(
                "Triggering is not supported for multiple events, the detector was "
                f"prepared with number_of_events={trigger_info.number_of_events}."
            )

        # Arm the detector and wait for it to finish.
        indices_written = await self._writer.get_indices_written()
        await self._controller.arm()
        await self._controller.wait_for_idle()
        end_observation = indices_written + 1

        async for index in self._writer.observe_indices_written(
            DEFAULT_TIMEOUT + (trigger_info.livetime or 0) + trigger_info.deadtime
        ):
            if index >= end_observation:
                break

    @AsyncStatus.wrap
    async def prepare(self, value: TriggerInfo) -> None:
        """Arm detector.

        Prepare the detector with trigger information. This is determined at and passed
        in from the plan level.

        :param value: TriggerInfo describing how to trigger the detector
        """
        if value.trigger != DetectorTrigger.INTERNAL and not value.deadtime:
            msg = "Deadtime must be supplied when in externally triggered mode"
            raise ValueError(msg)
        if not value.deadtime:
            value.deadtime = self._controller.get_deadtime(value.livetime)
        self._trigger_info = value
        self._number_of_events_iter = iter(
            value.number_of_events
            if isinstance(value.number_of_events, list)
            else [value.number_of_events]
        )

        await self._controller.prepare(value)
        self._describe = await self._writer.open(self.name, value.exposures_per_event)

        self._initial_frame = await self._writer.get_indices_written()
        if value.trigger != DetectorTrigger.INTERNAL:
            await self._controller.arm()
        self._trigger_info = value

    @AsyncStatus.wrap
    async def kickoff(self):
        if self._trigger_info is None or self._number_of_events_iter is None:
            raise RuntimeError("Prepare must be called before kickoff!")
        if self._trigger_info.trigger == DetectorTrigger.INTERNAL:
            await self._controller.arm()
        self._fly_start = time.monotonic()
        try:
            self._events_to_complete = next(self._number_of_events_iter)
            self._completable_exposures += (
                self._events_to_complete * self._trigger_info.exposures_per_event
            )
        except StopIteration as err:
            raise RuntimeError(
                f"Kickoff called more than the configured number of "
                f"{self._trigger_info.total_number_of_exposures} iteration(s)!"
            ) from err

    @WatchableAsyncStatus.wrap
    async def complete(self):
        trigger_info = _ensure_trigger_info_exists(self._trigger_info)
        indices_written = self._writer.observe_indices_written(
            trigger_info.exposure_timeout
            or (
                DEFAULT_TIMEOUT
                + (trigger_info.livetime or 0)
                + (trigger_info.deadtime or 0)
            )
        )
        try:
            async for index in indices_written:
                yield WatcherUpdate(
                    name=self.name,
                    current=index,
                    initial=self._initial_frame,
                    target=self._events_to_complete,
                    unit="",
                    precision=0,
                    time_elapsed=time.monotonic() - self._fly_start
                    if self._fly_start
                    else None,
                )
                if index >= self._events_to_complete:
                    break
        finally:
            await indices_written.aclose()
            if self._completable_exposures >= trigger_info.total_number_of_exposures:
                self._completable_exposures = 0
                self._events_to_complete = 0
                self._number_of_events_iter = None
                await self._controller.wait_for_idle()

    async def describe_collect(self) -> dict[str, DataKey]:
        return self._describe

    async def collect_asset_docs(
        self, index: int | None = None
    ) -> AsyncIterator[StreamAsset]:
        # Collect stream datum documents for all indices written.
        # The index is optional, and provided for fly scans, however this needs to be
        # retrieved for step scans.
        if index is None:
            index = await self._writer.get_indices_written()
        async for doc in self._writer.collect_stream_docs(self.name, index):
            yield doc

    async def get_index(self) -> int:
        return await self._writer.get_indices_written()

    @property
    def hints(self) -> Hints:
        return self._writer.get_hints(self.name)



import os
from collections.abc import Callable
from typing import cast

import pytest
from bluesky.run_engine import RunEngine

from ophyd_async.core import init_devices
from ophyd_async.core._providers import PathProvider
from ophyd_async.epics import adcore
from ophyd_async.testing import callback_on_mock_put, set_mock_value


@pytest.fixture
def ad_standard_det_factory(
    RE: RunEngine,
    static_path_provider_factory,
    static_filename_provider,
) -> Callable[
    [type[adcore.AreaDetector], type[adcore.ADWriter], int], adcore.AreaDetector
]:
    def generate_ad_standard_det(
        detector_cls: type[adcore.AreaDetector],
        writer_cls: type[adcore.ADWriter] = adcore.ADHDFWriter,
        path_provider: PathProvider | None = None,
        number=1,
        data_type=adcore.ADBaseDataType.UINT16,
        assume_file_path_exists: bool = False,
        **kwargs,
    ) -> adcore.AreaDetector:
        # Dynamically generate a name based on the class of controller
        detector_name = detector_cls.__name__
        if detector_name.endswith("Detector"):
            detector_name = detector_name[: -len("Detector")]

        # Create the static path provider here to allow for overriding the directory URI
        if path_provider is None:
            path_provider = static_path_provider_factory(static_filename_provider)

        with init_devices(mock=True):
            prefix = f"{detector_name.upper()}{number}:"
            name = f"test_ad{detector_name.lower()}{number}"

            test_adstandard_det = detector_cls(
                prefix,
                path_provider,
                # The inner areaDetector class wants the instantaiated object,
                # but the outward facing classes want a writer class type
                writer_cls=writer_cls,  # type: ignore
                name=name,
                **kwargs,
            )

        def on_set_file_path_callback(value: str, wait: bool = True):
            if os.path.exists(value) or assume_file_path_exists:
                set_mock_value(test_adstandard_det.fileio.file_path_exists, True)
                set_mock_value(
                    test_adstandard_det.fileio.full_file_name,
                    f"{value}/{static_filename_provider(device_name=test_adstandard_det.name)}{test_adstandard_det._writer._file_extension}",
                )

        callback_on_mock_put(
            test_adstandard_det.fileio.file_path, on_set_file_path_callback
        )

        # Set some sensible defaults to mimic a real detector setup
        set_mock_value(test_adstandard_det.driver.acquire_time, (number - 0.2))
        set_mock_value(test_adstandard_det.driver.acquire_period, float(number))
        set_mock_value(test_adstandard_det.fileio.capture, True)
        set_mock_value(test_adstandard_det.driver.data_type, data_type)

        # Set image mode to continuous to mimic a real detector setup
        set_mock_value(
            test_adstandard_det.driver.image_mode, adcore.ADImageMode.CONTINUOUS
        )

        if detector_cls == adcore.ContAcqAreaDetector:
            det = cast(adcore.ContAcqAreaDetector, test_adstandard_det)
            # For cont acq detectors, assume we are already acquiring
            set_mock_value(det.driver.acquire, True)

            # Also, setup the cb plugin with some reasonable values
            set_mock_value(det.cb_plugin.array_size_x, (9 + number))
            set_mock_value(det.cb_plugin.array_size_y, (9 + number))
            set_mock_value(det.cb_plugin.data_type, data_type)

        # Set number of frames per chunk and frame dimensions to something reasonable
        set_mock_value(test_adstandard_det.driver.array_size_x, (9 + number))
        set_mock_value(test_adstandard_det.driver.array_size_y, (9 + number))

        return test_adstandard_det

    return generate_ad_standard_det



from ophyd_async.core import (
    PathProvider,
    StandardDetector,
)
from ophyd_async.epics.odin import Odin, OdinWriter
from ophyd_async.fastcs.jungfrau._controller import JungfrauController
from ophyd_async.fastcs.jungfrau._signals import JungfrauDriverIO


class Jungfrau(StandardDetector[JungfrauController, OdinWriter]):
    """Ophyd-async implementation of a Jungfrau Detector."""

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider,
        drv_suffix: str,
        hdf_suffix: str,
        odin_nodes: int,
        name="",
    ):
        self.drv = JungfrauDriverIO(prefix + drv_suffix)
        self.odin = Odin(prefix + hdf_suffix, nodes=odin_nodes)
        writer = OdinWriter(
            path_provider,
            self.odin,
            self.drv.bit_depth,
        )
        controller = JungfrauController(self.drv)
        super().__init__(controller, writer, name=name)



from ophyd_async.core import (
    AsyncStatus,
    DetectorController,
    DetectorTrigger,
    TriggerInfo,
    wait_for_value,
)

from ._block import PcapBlock


class PandaPcapController(DetectorController):
    """For controlling a PCAP capture on the PandA."""

    def __init__(self, pcap: PcapBlock) -> None:
        self.pcap = pcap
        self._arm_status: AsyncStatus | None = None

    def get_deadtime(self, exposure: float | None) -> float:
        return 0.000000008

    async def prepare(self, trigger_info: TriggerInfo):
        if trigger_info.trigger not in (
            DetectorTrigger.CONSTANT_GATE,
            DetectorTrigger.VARIABLE_GATE,
        ):
            msg = (
                "Only constant_gate and variable_gate triggering is supported on "
                "the PandA",
            )
            raise TypeError(msg)

    async def arm(self):
        self._arm_status = self.pcap.arm.set(True)
        await wait_for_value(self.pcap.active, True, timeout=1)

    async def wait_for_idle(self):
        pass

    async def disarm(self):
        await self.pcap.arm.set(False)
        await wait_for_value(self.pcap.active, False, timeout=1)
        if self._arm_status and not self._arm_status.done:
            await self._arm_status
        self._arm_status = None



from __future__ import annotations

from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Protocol,
    TypeVar,
    runtime_checkable,
)

from bluesky.protocols import HasName, Location, Reading, T_co
from event_model import DataKey

from ._utils import T

if TYPE_CHECKING:
    from ._status import AsyncStatus


@runtime_checkable
class AsyncReadable(HasName, Protocol):
    """Async implementations of the sync [](#bluesky.protocols.Readable)."""

    @abstractmethod
    async def read(self) -> dict[str, Reading]:
        """Return value, timestamp, optional per-point metadata for each field name.

        For example:

            {
                "channel1": {"value": 5, "timestamp": 1472493713.271991},
                "channel2": {"value": 16, "timestamp": 1472493713.539238},
            }
        """

    @abstractmethod
    async def describe(self) -> dict[str, DataKey]:
        """Return per-scan metadata for each field name in `read()`.

        For example:

            {
                "channel1": {"source": "SOME_PV1", "dtype": "number", "shape": []},
                "channel2": {"source": "SOME_PV2", "dtype": "number", "shape": []},
            }
        """


@runtime_checkable
class AsyncConfigurable(HasName, Protocol):
    """Async implementation of the sync [](#bluesky.protocols.Configurable)."""

    @abstractmethod
    async def read_configuration(self) -> dict[str, Reading]:
        """Return value, timestamp, optional per-point metadata for each field name.

        Same API as [](#AsyncReadable.read) but for slow-changing fields related to
        configuration. e.g., exposure time. These will typically be read only
        once per run.
        """

    @abstractmethod
    async def describe_configuration(self) -> dict[str, DataKey]:
        """Return per-scan metadata for each field name in `read_configuration()`."""


@runtime_checkable
class AsyncPausable(Protocol):
    """Async implementation of the sync [](#bluesky.protocols.Pausable)."""

    @abstractmethod
    async def pause(self) -> None:
        """Perform device-specific work when the RunEngine pauses."""

    @abstractmethod
    async def resume(self) -> None:
        """Perform device-specific work when the RunEngine resumes after a pause."""


@runtime_checkable
class AsyncStageable(Protocol):
    """Async implementation of the sync [](#bluesky.protocols.Stageable)."""

    @abstractmethod
    def stage(self) -> AsyncStatus:
        """Set up the device for acquisition.

        :return: An `AsyncStatus` that is marked done when the device is done staging.
        """

    @abstractmethod
    def unstage(self) -> AsyncStatus:
        """Clean up the device after acquisition.

        :return: An `AsyncStatus` that is marked done when the device is done unstaging.
        """


@runtime_checkable
class AsyncMovable(Protocol[T_co]):
    @abstractmethod
    def set(self, value: T_co) -> AsyncStatus:
        """Return a ``Status`` that is marked done when the device is done moving."""


@runtime_checkable
class AsyncLocatable(AsyncMovable[T], Protocol):
    @abstractmethod
    async def locate(self) -> Location[T]:
        """Return the current location of a Device.

        While a ``Readable`` reports many values, a ``Movable`` will have the
        concept of location. This is where the Device currently is, and where it
        was last requested to move to. This protocol formalizes how to get the
        location from a ``Movable``.
        """


C = TypeVar("C", contravariant=True)


class Watcher(Protocol, Generic[C]):
    """Protocol for watching changes in values."""

    def __call__(
        self,
        current: C | None = None,
        initial: C | None = None,
        target: C | None = None,
        name: str | None = None,
        unit: str | None = None,
        precision: int | None = None,
        fraction: float | None = None,
        time_elapsed: float | None = None,
        time_remaining: float | None = None,
    ) -> Any: ...



import asyncio

from ophyd_async.core import (
    DetectorTrigger,
    TriggerInfo,
)
from ophyd_async.epics import adcore

from ._andor_io import Andor2DriverIO, Andor2TriggerMode

_MIN_DEAD_TIME = 0.1
_MAX_NUM_IMAGE = 999_999


# The deadtime of an Andor2 controller varies depending on the exact model of camera.
# Ideally we would maximize performance by dynamically retrieving the deadtime at
# runtime. See https://github.com/bluesky/ophyd-async/issues/308
class Andor2Controller(adcore.ADBaseController[Andor2DriverIO]):
    """DetectorCobntroller for Andor2DriverIO."""

    def get_deadtime(self, exposure: float | None) -> float:
        return _MIN_DEAD_TIME + (exposure or 0)

    async def prepare(self, trigger_info: TriggerInfo):
        await self.set_exposure_time_and_acquire_period_if_supplied(
            trigger_info.livetime
        )

        await asyncio.gather(
            self.driver.trigger_mode.set(self._get_trigger_mode(trigger_info.trigger)),
            self.driver.num_images.set(
                trigger_info.total_number_of_exposures or _MAX_NUM_IMAGE
            ),
            self.driver.image_mode.set(adcore.ADImageMode.MULTIPLE),
        )

    def _get_trigger_mode(self, trigger: DetectorTrigger) -> Andor2TriggerMode:
        supported_trigger_types = {
            DetectorTrigger.INTERNAL: Andor2TriggerMode.INTERNAL,
            DetectorTrigger.EDGE_TRIGGER: Andor2TriggerMode.EXT_TRIGGER,
        }
        if trigger not in supported_trigger_types:
            raise ValueError(
                f"{self.__class__.__name__} only supports the following trigger "
                f"types: {supported_trigger_types} but was asked to "
                f"use {trigger}"
            )
        return supported_trigger_types[trigger]



from ophyd_async.epics import adcore

from ._sim_io import SimDriverIO


class SimController(adcore.ADBaseController[SimDriverIO]):
    """Controller for simulated Areadetector."""

    def __init__(
        self,
        driver: SimDriverIO,
        good_states: frozenset[adcore.ADState] = adcore.DEFAULT_GOOD_STATES,
    ) -> None:
        super().__init__(driver, good_states=good_states)

    def get_deadtime(self, exposure: float | None) -> float:
        return 0.001



import asyncio
import logging

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    DetectorController,
    DetectorTrigger,
    TriggerInfo,
    wait_for_value,
)

from ._signals import (
    JUNGFRAU_TRIGGER_MODE_MAP,
    AcquisitionType,
    DetectorStatus,
    JungfrauDriverIO,
    PedestalMode,
)

# Deadtime is dependant on a wide combination of settings and on trigger mode
# but this is safe upper-limit
JUNGFRAU_DEADTIME_S = 2e-5

logger = logging.getLogger("ophyd_async")


class JungfrauController(DetectorController):
    def __init__(self, driver: JungfrauDriverIO):
        self._driver = driver

    def get_deadtime(self, exposure: float | None = None) -> float:
        return JUNGFRAU_DEADTIME_S

    async def prepare(self, trigger_info: TriggerInfo) -> None:
        # ValueErrors and warnings in this function come from the jungfrau operation
        # docs: https://rtd.xfel.eu/docs/jungfrau-detector-documentation/en/latest/operation.html

        # Deadtime here is really used as "time between frames"

        acquisition_type = await self._driver.acquisition_type.get_value()
        logger.info(f"Preparing Jungfrau in {acquisition_type} mode.")

        if trigger_info.trigger not in (
            DetectorTrigger.INTERNAL,
            DetectorTrigger.EDGE_TRIGGER,
        ):
            raise ValueError(
                "The trigger method can only be called with internal or edge triggering"
            )
        if (
            acquisition_type == AcquisitionType.PEDESTAL
            and trigger_info.trigger != DetectorTrigger.INTERNAL
        ):
            raise ValueError(
                "Jungfrau must be triggered internally while in pedestal mode."
            )

        if not isinstance(trigger_info.number_of_events, int):
            raise TypeError("Number of events must be an integer")

        if acquisition_type != AcquisitionType.PEDESTAL:
            if (
                trigger_info.trigger == DetectorTrigger.INTERNAL
                and trigger_info.number_of_events != 1
            ):
                raise ValueError(
                    "Number of events must be set to 1 in internal trigger mode during "
                    "standard acquisitions."
                )

            if (
                trigger_info.trigger == DetectorTrigger.EDGE_TRIGGER
                and trigger_info.exposures_per_event != 1
            ):
                raise ValueError(
                    "Exposures per event must be set to 1 in edge trigger mode "
                    "during standard acquisitions."
                )

        if not trigger_info.livetime:
            raise ValueError("Must set TriggerInfo.livetime")

        if trigger_info.livetime < 2e-6:
            logger.warning("Exposure time shorter than 2μs is not recommended")

        period_between_frames = trigger_info.livetime + trigger_info.deadtime

        if period_between_frames < self.get_deadtime():
            raise ValueError(
                f"Period between frames (exposure time - deadtime) = "
                f"{period_between_frames}s cannot be lower than minimum detector "
                f"deadtime {self.get_deadtime()}"
            )

        coros = [
            self._driver.trigger_mode.set(
                JUNGFRAU_TRIGGER_MODE_MAP[trigger_info.trigger]
            ),
            self._driver.period_between_frames.set(period_between_frames),
            self._driver.exposure_time.set(trigger_info.livetime),
        ]

        match acquisition_type:
            case AcquisitionType.STANDARD:
                frames_signal = (
                    trigger_info.exposures_per_event
                    if trigger_info.trigger is DetectorTrigger.INTERNAL
                    else trigger_info.number_of_events
                )
                coros.extend(
                    [
                        self._driver.frames_per_acq.set(frames_signal),
                    ]
                )
            case AcquisitionType.PEDESTAL:
                if trigger_info.number_of_events % 2 == 0:
                    coros.extend(
                        [
                            self._driver.pedestal_mode_frames.set(
                                trigger_info.exposures_per_event
                            ),
                            # No. events is double the pedestal loops,
                            # since pedestal scan does the entire loop
                            # twice.
                            self._driver.pedestal_mode_loops.set(
                                int(trigger_info.number_of_events / 2)
                            ),
                        ]
                    )
                else:
                    raise ValueError(
                        f"Invalid trigger info for pedestal mode. "
                        f"{trigger_info.number_of_events=} must be divisible by two. "
                        f"Was create_jungfrau_pedestal_triggering_info used?"
                    )

        await asyncio.gather(*coros)

        # Setting signals once the detector is in pedestal mode can cause errors,
        # so do this last
        if acquisition_type == AcquisitionType.PEDESTAL:
            await self._driver.pedestal_mode_state.set(PedestalMode.ON)

    async def arm(self):
        await self._driver.acquisition_start.trigger()

    async def wait_for_idle(self):
        await wait_for_value(
            self._driver.detector_status, DetectorStatus.IDLE, timeout=DEFAULT_TIMEOUT
        )

    async def disarm(self):
        await self._driver.acquisition_stop.trigger()
        await asyncio.gather(
            self._driver.pedestal_mode_state.set(PedestalMode.OFF),
            self._driver.acquisition_type.set(AcquisitionType.STANDARD),
        )



import asyncio
import functools
import logging
import time
from abc import abstractmethod
from collections.abc import Callable, Coroutine, Sequence
from typing import (
    Any,
    ParamSpec,
    TypeVar,
    cast,
    get_args,
    get_origin,
)

import numpy as np
import numpy.typing as npt
from bluesky.protocols import Reading
from event_model import DataKey, Limits, LimitsRange
from event_model.documents.event_descriptor import RdsRange
from tango import (
    AttrDataFormat,
    AttributeInfo,
    AttributeInfoEx,
    CmdArgType,
    CommandInfo,
    DevFailed,  # type: ignore
    DeviceProxy,
    DevState,
    EventType,
    GreenMode,
)
from tango.asyncio import DeviceProxy as AsyncDeviceProxy
from tango.asyncio_executor import (
    AsyncioExecutor,
    get_global_executor,
    set_global_executor,
)
from tango.utils import is_binary, is_bool, is_float, is_int, is_str

from ophyd_async.core import (
    Array1D,
    AsyncStatus,
    Callback,
    NotConnectedError,
    SignalBackend,
    SignalDatatypeT,
    SignalMetadata,
    StrictEnum,
    Table,
    get_dtype,
    make_datakey,
    wait_for_connection,
)
from ophyd_async.tango.testing import TestConfig

from ._converters import (
    TangoConverter,
    TangoDevStateArrayConverter,
    TangoDevStateConverter,
    TangoEnumArrayConverter,
    TangoEnumConverter,
)
from ._utils import DevStateEnum, get_device_trl_and_attr, try_to_cast_as_float

logger = logging.getLogger("ophyd_async")

# time constant to wait for timeout
A_BIT = 1e-5

P = ParamSpec("P")
R = TypeVar("R")


def ensure_proper_executor(
    func: Callable[P, Coroutine[Any, Any, R]],
) -> Callable[P, Coroutine[Any, Any, R]]:
    """Ensure decorated method has a proper asyncio executor."""

    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        current_executor: AsyncioExecutor = get_global_executor()  # type: ignore
        if not current_executor.in_executor_context():  # type: ignore
            set_global_executor(AsyncioExecutor())
        return await func(*args, **kwargs)

    return wrapper


class TangoLongStringTable(Table):
    long: Array1D[np.int32]
    string: Sequence[str]


class TangoDoubleStringTable(Table):
    double: Array1D[np.float64]
    string: Sequence[str]


def get_python_type(config: AttributeInfoEx | CommandInfo | TestConfig) -> object:
    """For converting between recieved tango types and python primatives."""
    tango_type = None
    tango_format = None
    if isinstance(config, AttributeInfoEx | AttributeInfo):
        tango_type = config.data_type
        tango_format = config.data_format
    elif isinstance(config, CommandInfo):
        read_character = get_command_character(config)
        if read_character == CommandProxyReadCharacter.READ:
            tango_type = config.out_type
        else:
            tango_type = config.in_type
    elif isinstance(config, TestConfig):
        tango_type = config.data_type
        tango_format = config.data_format
    else:
        raise TypeError("Unrecognized Tango resource configuration")
    if tango_format not in [
        AttrDataFormat.SCALAR,
        AttrDataFormat.SPECTRUM,
        AttrDataFormat.IMAGE,
        None,
    ]:
        raise TypeError("Unknown TangoFormat")

    if tango_type is CmdArgType.DevVarLongStringArray:
        return TangoLongStringTable
    if tango_type is CmdArgType.DevVarDoubleStringArray:
        return TangoDoubleStringTable

    def _get_type(cls: type) -> object:
        if tango_format == AttrDataFormat.SCALAR:
            return cls
        elif tango_format == AttrDataFormat.SPECTRUM:
            if cls is str or issubclass(cls, StrictEnum):
                return Sequence[cls]
            return Array1D[cls]
        elif tango_format == AttrDataFormat.IMAGE:
            if cls is str or issubclass(cls, StrictEnum):
                return Sequence[Sequence[str]]
            return npt.NDArray[cls]
        else:
            return cls

    if is_int(tango_type, True):
        return _get_type(int)
    elif is_float(tango_type, True):
        return _get_type(float)
    elif is_bool(tango_type, True):
        return _get_type(bool)
    elif is_str(tango_type, True):
        return _get_type(str)
    elif is_binary(tango_type, True):
        return _get_type(str)
    elif tango_type == CmdArgType.DevEnum:
        if hasattr(config, "enum_labels"):
            enum_dict = {label: str(label) for label in config.enum_labels}
            return _get_type(StrictEnum("TangoEnum", enum_dict))
        else:
            return _get_type(int)
    elif tango_type == CmdArgType.DevState:
        return _get_type(DevStateEnum)
    elif tango_type == CmdArgType.DevUChar:
        return _get_type(int)
    elif tango_type == CmdArgType.DevVoid:
        return None
    else:
        raise TypeError(f"Unknown TangoType: {tango_type}")


class TangoProxy:
    support_events: bool = True
    _proxy: DeviceProxy
    _name: str
    _converter: TangoConverter = TangoConverter()

    def __init__(self, device_proxy: DeviceProxy, name: str):
        self._proxy = device_proxy
        self._name = name

    async def connect(self) -> None:
        """Perform actions after proxy is connected.

        e.g. check if signal can be subscribed.
        """

    @abstractmethod
    async def get(self) -> object:
        """Get value from TRL."""

    @abstractmethod
    async def get_w_value(self) -> object:
        """Get last written value from TRL."""

    @abstractmethod
    async def put(
        self, value: object | None, wait: bool = True, timeout: float | None = None
    ) -> AsyncStatus | None:
        """Put value to TRL."""

    @abstractmethod
    async def get_config(self) -> AttributeInfoEx | CommandInfo:
        """Get TRL config async."""

    @abstractmethod
    async def get_reading(self) -> Reading:
        """Get reading from TRL."""

    @abstractmethod
    def has_subscription(self) -> bool:
        """Indicate that this trl already subscribed."""

    @abstractmethod
    def subscribe_callback(self, callback: Callback | None):
        """Subscribe tango CHANGE event to callback."""

    @abstractmethod
    def unsubscribe_callback(self):
        """Delete CHANGE event subscription."""

    @abstractmethod
    def set_polling(
        self,
        allow_polling: bool = True,
        polling_period: float = 0.1,
        abs_change=None,
        rel_change=None,
    ):
        """Set polling parameters."""

    def set_converter(self, converter: "TangoConverter"):
        self._converter = converter


class AttributeProxy(TangoProxy):
    """Used by the tango transport."""

    _callback: Callback | None = None
    _eid: int | None = None
    _poll_task: asyncio.Task | None = None
    _abs_change: float | None = None
    _rel_change: float | None = 0.1
    _polling_period: float = 0.1
    _allow_polling: bool = False
    exception: BaseException | None = None
    _last_reading: Reading = Reading(value=None, timestamp=0, alarm_severity=0)

    async def connect(self) -> None:
        try:
            # I have to typehint proxy as tango.DeviceProxy because
            # tango.asyncio.DeviceProxy cannot be used as a typehint.
            # This means pyright will not be able to see that
            # subscribe_event is awaitable.
            eid = await self._proxy.subscribe_event(  # type: ignore
                self._name, EventType.CHANGE_EVENT, self._event_processor
            )
            await self._proxy.unsubscribe_event(eid)
            self.support_events = True
        except Exception:
            pass

    @ensure_proper_executor
    async def get(self) -> object:  # type: ignore
        attr = await self._proxy.read_attribute(self._name)
        return self._converter.value(attr.value)

    @ensure_proper_executor
    async def get_w_value(self) -> object:  # type: ignore
        attr = await self._proxy.read_attribute(self._name)
        return self._converter.value(attr.w_value)

    @ensure_proper_executor
    async def put(  # type: ignore
        self, value: object | None, wait: bool = True, timeout: float | None = None
    ) -> AsyncStatus | None:
        if wait is False:
            raise RuntimeWarning(
                "wait=False is not supported in Tango."
                "Simply don't await the status object."
            )
        # TODO: remove the timeout from this as it is handled at the signal level
        value = self._converter.write_value(value)
        try:

            async def _write():
                return await self._proxy.write_attribute(self._name, value)

            task = asyncio.create_task(_write())
            await asyncio.wait_for(task, timeout)
        except TimeoutError as te:
            raise TimeoutError(f"{self._name} attr put failed: Timeout") from te
        except DevFailed as de:
            raise RuntimeError(
                f"{self._name} device failure: {de.args[0].desc}"
            ) from de

    @ensure_proper_executor
    async def get_config(self) -> AttributeInfoEx:  # type: ignore
        return await self._proxy.get_attribute_config(self._name)

    @ensure_proper_executor
    async def get_reading(self) -> Reading:  # type: ignore
        attr = await self._proxy.read_attribute(self._name)
        reading = Reading(
            value=self._converter.value(attr.value),
            timestamp=attr.time.totime(),
            alarm_severity=attr.quality,
        )
        self._last_reading = reading
        return reading

    def has_subscription(self) -> bool:
        return bool(self._callback)

    @ensure_proper_executor
    async def _subscribe_to_event(self):
        if not self._eid:
            self._eid = await self._proxy.subscribe_event(
                self._name,
                EventType.CHANGE_EVENT,
                self._event_processor,
                stateless=True,
                green_mode=GreenMode.Asyncio,
            )

    def subscribe_callback(self, callback: Callback | None):
        # If the attribute supports events, then we can subscribe to them
        # If the callback is not a callable, then we raise an error
        if callback is not None and not callable(callback):
            raise RuntimeError("Callback must be a callable")

        self._callback = callback
        if self.support_events:
            asyncio.create_task(self._subscribe_to_event())
        elif self._allow_polling:
            """start polling if no events supported"""
            if self._callback is not None:

                async def _poll():
                    while True:
                        try:
                            await self.poll()
                        except RuntimeError as exc:
                            self.exception = exc
                            await asyncio.sleep(1)

                self._poll_task = asyncio.create_task(_poll())
        else:
            self.unsubscribe_callback()
            raise RuntimeError(
                f"Cannot set event for {self._name}. "
                "Cannot set a callback on an attribute that does not support events and"
                " for which polling is disabled."
            )

    def unsubscribe_callback(self):
        if self._eid:
            try:
                self._proxy.unsubscribe_event(self._eid, green_mode=False)
            except Exception as exc:
                logger.warning(f"Could not unsubscribe from event: {exc}")
            finally:
                self._eid = None
        if self._poll_task:
            self._poll_task.cancel()
            self._poll_task = None
            if self._callback is not None:
                # Call the callback with the last reading
                try:
                    self._callback(self._last_reading)
                except TypeError:
                    pass
        self._callback = None

    @ensure_proper_executor
    async def _event_processor(self, event):
        if not event.err:
            reading = Reading(
                value=self._converter.value(event.attr_value.value),
                timestamp=event.get_date().totime(),
                alarm_severity=event.attr_value.quality,
            )
            if self._callback is not None:
                self._callback(reading)

    async def poll(self):
        """Poll the attribute and call the callback if the value has changed.

        Only callback if value has changed by more than the absolute or relative
        change. This function is used when an attribute that does not support
        events is cached or a callback is passed to it.
        """
        try:
            last_reading = await self.get_reading()
            flag = 0
            # Initial reading
            if self._callback is not None:
                self._callback(last_reading)
        except Exception as exc:
            raise RuntimeError(f"Could not poll the attribute: {exc}") from exc

        try:
            # If the value is a number, we can check for changes
            if isinstance(last_reading["value"], int | float):
                while True:
                    await asyncio.sleep(self._polling_period)
                    reading = await self.get_reading()
                    if reading is None or reading["value"] is None:
                        continue
                    diff = abs(reading["value"] - last_reading["value"])
                    if self._abs_change is not None and diff >= abs(self._abs_change):
                        if self._callback is not None:
                            self._callback(reading)
                            flag = 0

                    elif (
                        self._rel_change is not None
                        and diff >= self._rel_change * abs(last_reading["value"])
                    ):
                        if self._callback is not None:
                            self._callback(reading)
                            flag = 0

                    else:
                        flag = (flag + 1) % 4
                        if flag == 0 and self._callback is not None:
                            self._callback(reading)

                    last_reading = reading.copy()
                    if self._callback is None:
                        break
            # If the value is not a number, we can only poll
            else:
                while True:
                    await asyncio.sleep(self._polling_period)
                    flag = (flag + 1) % 4
                    if flag == 0:
                        reading = await self.get_reading()
                        if reading is None or reading["value"] is None:
                            continue
                        if isinstance(reading["value"], np.ndarray):
                            if not np.array_equal(
                                reading["value"], last_reading["value"]
                            ):
                                if self._callback is not None:
                                    self._callback(reading)
                                else:
                                    break
                        else:
                            if reading["value"] != last_reading["value"]:
                                if self._callback is not None:
                                    self._callback(reading)
                                else:
                                    break
                        last_reading = reading.copy()
        except Exception as exc:
            raise RuntimeError(f"Could not poll the attribute: {exc}") from exc

    def set_polling(
        self,
        allow_polling: bool = False,
        polling_period: float = 0.5,
        abs_change: float | None = None,
        rel_change: float | None = 0.1,
    ):
        """Set the polling parameters."""
        self._allow_polling = allow_polling
        self._polling_period = polling_period
        self._abs_change = abs_change
        self._rel_change = rel_change


class CommandProxyReadCharacter(StrictEnum):
    """Enum to carry the read/write character of the CommandProxy."""

    READ = "READ"
    WRITE = "WRITE"
    READ_WRITE = "READ_WRITE"
    EXECUTE = "EXECUTE"


def get_command_character(config: CommandInfo) -> CommandProxyReadCharacter:
    """Return the command character for the given command config."""
    in_type = config.in_type
    out_type = config.out_type
    if in_type == CmdArgType.DevVoid and out_type != CmdArgType.DevVoid:
        read_character = CommandProxyReadCharacter.READ
    elif in_type != CmdArgType.DevVoid and out_type == CmdArgType.DevVoid:
        read_character = CommandProxyReadCharacter.WRITE
    elif in_type == CmdArgType.DevVoid and out_type == CmdArgType.DevVoid:
        read_character = CommandProxyReadCharacter.EXECUTE
    else:
        read_character = CommandProxyReadCharacter.READ_WRITE
    return read_character


class CommandProxy(TangoProxy):
    """Tango proxy for commands."""

    _last_reading: Reading
    _last_w_value: Any
    _config: CommandInfo
    _read_character: CommandProxyReadCharacter
    device_proxy: DeviceProxy
    name: str

    def __init__(self, device_proxy: DeviceProxy, name: str):
        super().__init__(device_proxy, name)
        self._last_reading = Reading(value=None, timestamp=0, alarm_severity=0)
        self.device_proxy = device_proxy
        self.name = name
        self._last_w_value = None

    def subscribe_callback(self, callback: Callback | None) -> None:
        raise NotImplementedError("Cannot subscribe to commands")

    def unsubscribe_callback(self) -> None:
        raise NotImplementedError("Cannot unsubscribe from commands")

    async def get(self) -> object:
        if self._read_character == CommandProxyReadCharacter.READ_WRITE:
            return self._last_reading["value"]
        elif self._read_character == CommandProxyReadCharacter.READ:
            await self.put(value=None, wait=True, timeout=None)
            return self._last_reading["value"]

    async def get_w_value(self) -> object:
        return self._last_w_value

    async def connect(self) -> None:
        self._config = await self.device_proxy.get_command_config(self.name)
        self._read_character = get_command_character(self._config)

    @ensure_proper_executor
    async def put(  # type: ignore
        self, value: object | None, wait: bool = True, timeout: float | None = None
    ) -> AsyncStatus | None:
        if wait is False:
            raise RuntimeError(
                "wait=False is not supported in Tango."
                " Simply don't await the status object."
            )
        value = self._converter.write_value(value)
        try:

            async def _put():
                return await self._proxy.command_inout(self._name, value)

            task = asyncio.create_task(_put())
            val = await asyncio.wait_for(task, timeout)
            self._last_w_value = value
            self._last_reading = Reading(
                value=self._converter.value(val),
                timestamp=time.time(),
                alarm_severity=0,
            )
        except TimeoutError as te:
            raise TimeoutError(f"{self._name} command failed: Timeout") from te
        except DevFailed as de:
            raise RuntimeError(
                f"{self._name} device failure: {de.args[0].desc}"
            ) from de

    @ensure_proper_executor
    async def get_config(self) -> CommandInfo:  # type: ignore
        return await self._proxy.get_command_config(self._name)

    async def get_reading(self) -> Reading:
        if self._read_character == CommandProxyReadCharacter.READ:
            await self.put(value=None, wait=True, timeout=None)
            return self._last_reading
        else:
            return self._last_reading

    def set_polling(
        self,
        allow_polling: bool = False,
        polling_period: float = 0.5,
        abs_change: float | None = None,
        rel_change: float | None = 0.1,
    ):
        pass


def get_dtype_extended(datatype) -> object | None:
    """For converting tango types to numpy datatype formats."""
    # DevState tango type does not have numpy equivalents
    dtype = get_dtype(datatype)
    if dtype == np.object_:
        if datatype.__args__[1].__args__[0] in [DevStateEnum, DevState]:
            dtype = CmdArgType.DevState
    return dtype


def get_source_metadata(
    tango_resource: str,
    tr_configs: dict[str, AttributeInfoEx],
) -> SignalMetadata:
    metadata = {}
    for _, config in tr_configs.items():
        if isinstance(config, AttributeInfoEx):
            alarm_info = config.alarms
            _limits = Limits(
                control=LimitsRange(
                    low=try_to_cast_as_float(config.min_value),
                    high=try_to_cast_as_float(config.max_value),
                ),
                warning=LimitsRange(
                    low=try_to_cast_as_float(alarm_info.min_warning),
                    high=try_to_cast_as_float(alarm_info.max_warning),
                ),
                alarm=LimitsRange(
                    low=try_to_cast_as_float(alarm_info.min_alarm),
                    high=try_to_cast_as_float(alarm_info.max_alarm),
                ),
            )

            delta_t, delta_val = map(
                try_to_cast_as_float, (alarm_info.delta_t, alarm_info.delta_val)
            )
            if isinstance(delta_t, float) and isinstance(delta_val, float):
                limits_rds = RdsRange(
                    time_difference=delta_t,
                    value_difference=delta_val,
                )
                _limits["rds"] = limits_rds
            # if only one of the two is set
            elif isinstance(delta_t, float) ^ isinstance(delta_val, float):
                logger.warning(
                    f"Both delta_t and delta_val should be set for {tango_resource} "
                    f"but only one is set. "
                    f"delta_t: {alarm_info.delta_t}, delta_val: {alarm_info.delta_val}"
                )

            _choices = list(config.enum_labels) if config.enum_labels else []

            tr_dtype = get_python_type(config)

            if tr_dtype == CmdArgType.DevState:
                _choices = list(DevState.names.keys())

            _precision = None
            if config.format:
                try:
                    _precision = int(config.format.split(".")[1].split("f")[0])
                except (ValueError, IndexError) as exc:
                    # If parsing config.format fails, _precision remains None.
                    logger.warning(
                        "Failed to parse precision from config.format: %s. Error: %s",
                        config.format,
                        exc,
                    )
            no_limits = Limits(
                control=LimitsRange(high=None, low=None),
                warning=LimitsRange(high=None, low=None),
                alarm=LimitsRange(high=None, low=None),
            )
            if _limits:
                if _limits != no_limits:
                    metadata["limits"] = _limits
            if _choices:
                metadata["choices"] = _choices
            if _precision:
                metadata["precision"] = _precision
            if config.unit:
                metadata["units"] = config.unit
    return SignalMetadata(**metadata)


async def get_tango_trl(
    full_trl: str, device_proxy: DeviceProxy | TangoProxy | None, timeout: float
) -> TangoProxy:
    """Get the tango resource locator."""
    if isinstance(device_proxy, TangoProxy):
        return device_proxy
    device_trl, trl_name = get_device_trl_and_attr(full_trl)
    trl_name = trl_name.lower()
    if device_proxy is None:
        device_proxy = await AsyncDeviceProxy(device_trl, timeout=timeout)
    # all attributes can be always accessible with low register
    if isinstance(device_proxy, DeviceProxy):
        all_attrs = [
            attr_name.lower() for attr_name in device_proxy.get_attribute_list()
        ]
    else:
        raise TypeError(
            f"device_proxy must be an instance of DeviceProxy for {full_trl}"
        )
    if trl_name in all_attrs:
        return AttributeProxy(device_proxy, trl_name)

    # all commands can be always accessible with low register
    all_cmds = [cmd_name.lower() for cmd_name in device_proxy.get_command_list()]
    if trl_name in all_cmds:
        return CommandProxy(device_proxy, trl_name)

    raise RuntimeError(f"{trl_name} cannot be found in {device_proxy.name()}")


def make_converter(info: AttributeInfoEx | CommandInfo, datatype) -> TangoConverter:
    if isinstance(info, AttributeInfoEx):
        match info.data_type:
            case CmdArgType.DevEnum:
                if datatype and issubclass(datatype, StrictEnum):
                    labels = [e.value for e in datatype]
                else:  # get from enum_labels metadata
                    labels = list(info.enum_labels)
                if info.data_format == AttrDataFormat.SCALAR:
                    return TangoEnumConverter(labels)
                elif info.data_format in [
                    AttrDataFormat.SPECTRUM,
                    AttrDataFormat.IMAGE,
                ]:
                    return TangoEnumArrayConverter(labels)
            case CmdArgType.DevState:
                if info.data_format == AttrDataFormat.SCALAR:
                    return TangoDevStateConverter()
                elif info.data_format in [
                    AttrDataFormat.SPECTRUM,
                    AttrDataFormat.IMAGE,
                ]:
                    return TangoDevStateArrayConverter()
    else:  # command info
        match info.in_type:
            case CmdArgType.DevState:
                return TangoDevStateConverter()
            case CmdArgType.DevEnum:
                if datatype and issubclass(datatype, StrictEnum):
                    labels = [e.value for e in datatype]
                    return TangoEnumConverter(labels)
                else:
                    logger.warning(
                        "No override enum class provided for Tango enum command"
                    )
    # default case return trivial converter
    return TangoConverter()


class TangoSignalBackend(SignalBackend[SignalDatatypeT]):
    """Tango backend to connect signals over tango."""

    def __init__(
        self,
        datatype: type[SignalDatatypeT] | None,
        read_trl: str = "",
        write_trl: str = "",
        device_proxy: DeviceProxy | None = None,
    ):
        self.device_proxy = device_proxy
        self.read_trl = read_trl
        self.write_trl = write_trl
        self.proxies: dict[str, TangoProxy | DeviceProxy | None] = {
            read_trl: self.device_proxy,
            write_trl: self.device_proxy,
        }
        self.trl_configs: dict[str, AttributeInfoEx] = {}
        self._polling: tuple[bool, float, float | None, float | None] = (
            False,
            0.1,
            None,
            0.1,
        )
        self.support_events: bool = True
        self.status: AsyncStatus | None = None
        self.converter = TangoConverter()  # gets replaced at connect
        super().__init__(datatype)

    @classmethod
    def datatype_allowed(cls, dtype: Any) -> bool:
        return dtype in (int, float, str, bool, np.ndarray, StrictEnum)

    def set_trl(self, read_trl: str = "", write_trl: str = ""):
        self.read_trl = read_trl
        self.write_trl = write_trl if write_trl else read_trl
        self.proxies = {
            read_trl: self.device_proxy,
            write_trl: self.device_proxy,
        }

    def source(self, name: str, read: bool) -> str:
        return self.read_trl if read else self.write_trl

    def _type_match_ndarray(self, signal_type: type[SignalDatatypeT], tr_dtype: object):
        tango_resource = self.source(name="", read=True)

        def extract_dtype_param(dtype_arg):
            if hasattr(dtype_arg, "__origin__") and dtype_arg.__origin__ is np.dtype:
                inner = get_args(dtype_arg)
                return inner[0] if inner else object
            return dtype_arg

        signal_dtype = extract_dtype_param(get_args(signal_type)[-1])
        tr_dtype_arg = extract_dtype_param(get_args(tr_dtype)[-1])

        try:
            sdt = np.dtype(signal_dtype)
            tdt = np.dtype(tr_dtype_arg)
        except TypeError as exc:
            raise TypeError(
                f"Could not interpret array dtypes: {signal_dtype!r},"
                f" {tr_dtype_arg!r} ({exc})"
            ) from exc

        if sdt != tdt:
            raise TypeError(
                f"{tango_resource} has type {tr_dtype!r}, expected {self.datatype!r}"
            )

    def _type_match_array(
        self,
        signal_type: type[SignalDatatypeT] | None,
        tr_dtype: object,
        tango_resource: str,
    ):
        # Always get a fresh resource string for the error context
        tango_resource = self.source(name="", read=True)
        if get_origin(signal_type) is Sequence and get_origin(tr_dtype) is Sequence:
            sig_elem_type = get_args(signal_type)[0]
            tr_elem_type = get_args(tr_dtype)[0]
            self._type_match_scalar(sig_elem_type, tr_elem_type, tango_resource)
            return
        elif (
            get_origin(signal_type) is np.ndarray and get_origin(tr_dtype) is np.ndarray
        ):
            if signal_type is None:
                raise TypeError(
                    f"{tango_resource} has type {tr_dtype!r}, expected a non-None"
                    f" signal_type"
                )
            self._type_match_ndarray(signal_type, tr_dtype)
            return
        else:
            raise TypeError(
                tango_resource, "has type", str(signal_type), "which is not recognized"
            )

    def _type_match_scalar(
        self,
        signal_type: type[SignalDatatypeT] | None,
        tr_dtype: object,
        tango_resource: str,
    ):
        if signal_type is tr_dtype:
            return
        if isinstance(signal_type, type) and issubclass(signal_type, StrictEnum):
            return
        raise TypeError(
            f"{tango_resource} has type {tr_dtype!r}, expected {self.datatype!r}"
        )

    def _verify_datatype_matches(self, config: AttributeInfoEx | CommandInfo):
        """Verify that the datatype of the config matches the datatype of the signal."""
        tr_dtype = get_python_type(config)
        tango_resource = self.source(name="", read=True)
        signal_type = self.datatype
        if isinstance(config, AttributeInfoEx | AttributeInfo):
            tr_format = config.data_format
            if tr_format in [AttrDataFormat.SPECTRUM, AttrDataFormat.IMAGE]:
                self._type_match_array(signal_type, tr_dtype, tango_resource)
            elif tr_format is AttrDataFormat.SCALAR:
                self._type_match_scalar(signal_type, tr_dtype, tango_resource)
        elif isinstance(config, CommandInfo):
            if (
                config.in_type != CmdArgType.DevVoid
                and config.out_type != CmdArgType.DevVoid
                and config.in_type != config.out_type
            ):
                raise RuntimeError(
                    "Commands with different in and out dtypes are not supported"
                )
            if get_origin(tr_dtype) in [Sequence, np.ndarray]:
                self._type_match_array(signal_type, tr_dtype, tango_resource)
            else:
                self._type_match_scalar(signal_type, tr_dtype, tango_resource)
        else:
            raise TypeError(
                f"Unrecognized resource configuration: {config} "
                f"for source {tango_resource}"
            )

    async def _connect_and_store_config(self, trl: str, timeout: float) -> None:
        if not trl:
            raise RuntimeError(f"trl not set for {self}")
        try:
            self.proxies[trl] = await get_tango_trl(trl, self.proxies[trl], timeout)
            if self.proxies[trl] is None:
                raise NotConnectedError(f"Not connected to {trl}")
            # Pyright does not believe that self.proxies[trl] is not None despite
            # the check above
            await self.proxies[trl].connect()  # type: ignore
            config = await self.proxies[trl].get_config()  # type: ignore
            self.trl_configs[trl] = config

            # Perform signal verification
            self._verify_datatype_matches(config)

            if isinstance(config, AttributeInfoEx):
                if (
                    config.data_type == CmdArgType.DevString
                    and config.data_format == AttrDataFormat.IMAGE
                ):
                    raise TypeError(
                        "DevString IMAGE attributes are not supported by ophyd-async."
                    )
            self.proxies[trl].support_events = self.support_events  # type: ignore
        except TimeoutError as ce:
            raise NotConnectedError(f"tango://{trl}") from ce

    async def connect(self, timeout: float) -> None:
        if not self.read_trl:
            raise RuntimeError(f"trl not set for {self}")
        if self.read_trl != self.write_trl:
            # Different, need to connect both
            await wait_for_connection(
                read_trl=self._connect_and_store_config(self.read_trl, timeout),
                write_trl=self._connect_and_store_config(self.write_trl, timeout),
            )
        else:
            # The same, so only need to connect one
            await self._connect_and_store_config(self.read_trl, timeout)
        self.proxies[self.read_trl].set_polling(*self._polling)  # type: ignore
        self.converter = make_converter(self.trl_configs[self.read_trl], self.datatype)
        self.proxies[self.read_trl].set_converter(self.converter)  # type: ignore

    async def put(self, value: SignalDatatypeT | None, wait=True, timeout=None) -> None:
        if self.proxies[self.write_trl] is None:
            raise NotConnectedError(f"Not connected to {self.write_trl}")
        self.status = None
        put_status = await self.proxies[self.write_trl].put(value, wait, timeout)  # type: ignore
        self.status = put_status

    async def get_datakey(self, source: str) -> DataKey:
        try:
            value: Any = await self.proxies[source].get()  # type: ignore
        except AttributeError as ae:
            raise NotConnectedError(f"Not connected to {source}") from ae
        md = get_source_metadata(source, self.trl_configs)
        return make_datakey(
            self.datatype,  # type: ignore
            value,
            source,
            metadata=md,
        )

    async def get_reading(self) -> Reading[SignalDatatypeT]:
        if self.proxies[self.read_trl] is None:
            raise NotConnectedError(f"Not connected to {self.read_trl}")
        reading = await self.proxies[self.read_trl].get_reading()  # type: ignore
        return reading

    async def get_value(self) -> SignalDatatypeT:
        if self.proxies[self.read_trl] is None:
            raise NotConnectedError(f"Not connected to {self.read_trl}")
        proxy = self.proxies[self.read_trl]
        if proxy is None:
            raise NotConnectedError(f"Not connected to {self.read_trl}")
        value = await proxy.get()
        return cast(SignalDatatypeT, value)

    async def get_setpoint(self) -> SignalDatatypeT:
        if self.proxies[self.write_trl] is None:
            raise NotConnectedError(f"Not connected to {self.write_trl}")
        proxy = self.proxies[self.write_trl]
        if proxy is None:
            raise NotConnectedError(f"Not connected to {self.write_trl}")
        w_value = await proxy.get_w_value()
        return cast(SignalDatatypeT, w_value)

    def set_callback(self, callback: Callback | None) -> None:
        if self.proxies[self.read_trl] is None:
            raise NotConnectedError(f"Not connected to {self.read_trl}")
        if self.support_events is False and self._polling[0] is False:
            raise RuntimeError(
                f"Cannot set event for {self.read_trl}. "
                "Cannot set a callback on an attribute that does not support events and"
                " for which polling is disabled."
            )

        if callback and self.proxies[self.read_trl].has_subscription():  # type: ignore
            msg = "Cannot set a callback when one is already set"
            raise RuntimeError(msg)

        if self.proxies[self.read_trl].has_subscription():  # type: ignore
            self.proxies[self.read_trl].unsubscribe_callback()  # type: ignore

        if callback:
            try:
                self.proxies[self.read_trl].subscribe_callback(callback)  # type: ignore
            except RuntimeError as exc:
                raise RuntimeError(
                    f"Cannot set callback for {self.read_trl}. {exc}"
                ) from exc

    def set_polling(
        self,
        allow_polling: bool = True,
        polling_period: float = 0.1,
        abs_change: float | None = None,
        rel_change: float | None = 0.1,
    ):
        self._polling = (allow_polling, polling_period, abs_change, rel_change)

    def allow_events(self, allow: bool = True):
        self.support_events = allow



import asyncio
from typing import Annotated as A

import numpy as np
from bluesky.protocols import Movable, Stoppable

from ophyd_async.core import (
    CALCULATE_TIMEOUT,
    DEFAULT_TIMEOUT,
    CalculatableTimeout,
    SignalR,
    SignalRW,
    SignalX,
    StandardReadable,
    WatchableAsyncStatus,
    WatcherUpdate,
    observe_value,
)
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.epics.core import EpicsDevice, PvSuffix


class DemoMotor(EpicsDevice, StandardReadable, Movable, Stoppable):
    """A demo movable that moves based on velocity."""

    # Whether set() should complete successfully or not
    _set_success = True
    # Define some signals
    readback: A[SignalR[float], PvSuffix("Readback"), Format.HINTED_SIGNAL]
    velocity: A[SignalRW[float], PvSuffix("Velocity"), Format.CONFIG_SIGNAL]
    units: A[SignalR[str], PvSuffix("Readback.EGU"), Format.CONFIG_SIGNAL]
    setpoint: A[SignalRW[float], PvSuffix("Setpoint")]
    precision: A[SignalR[int], PvSuffix("Readback.PREC")]
    # If a signal name clashes with a bluesky verb add _ to the attribute name
    stop_: A[SignalX, PvSuffix("Stop.PROC")]

    def set_name(self, name: str, *, child_name_separator: str | None = None) -> None:
        super().set_name(name, child_name_separator=child_name_separator)
        # Readback should be named the same as its parent in read()
        self.readback.set_name(name)

    @WatchableAsyncStatus.wrap
    async def set(  # type: ignore
        self, new_position: float, timeout: CalculatableTimeout = CALCULATE_TIMEOUT
    ):
        # The move should complete successfully unless stop(success=False) is called
        self._set_success = True
        # Get some variables for the progress bar reporting
        old_position, units, precision, velocity = await asyncio.gather(
            self.setpoint.get_value(),
            self.units.get_value(),
            self.precision.get_value(),
            self.velocity.get_value(),
        )
        # If not supplied, calculate a suitable timeout for the move
        if timeout == CALCULATE_TIMEOUT:
            timeout = abs(new_position - old_position) / velocity + DEFAULT_TIMEOUT
        # Wait for the value to set, but don't wait for put completion callback
        await self.setpoint.set(new_position, wait=False)
        # Observe the readback Signal, and on each new position...
        async for current_position in observe_value(
            self.readback, done_timeout=timeout
        ):
            # Emit a progress bar update
            yield WatcherUpdate(
                current=current_position,
                initial=old_position,
                target=new_position,
                name=self.name,
                unit=units,
                precision=precision,
            )
            # If we are at the desired position the break
            if np.isclose(current_position, new_position):
                break
        # If we were told to stop and report an error then do so
        if not self._set_success:
            raise RuntimeError("Motor was stopped")

    async def stop(self, success=True):
        self._set_success = success
        await self.stop_.trigger()



import itertools
import os
from pathlib import Path, PurePosixPath, PureWindowsPath
from typing import Any

import pytest

from ophyd_async.core import StaticPathProvider, TriggerInfo
from ophyd_async.epics import (
    adandor,
    adaravis,
    adcore,
    adkinetix,
    adpilatus,
    adsimdetector,
    advimba,
)

DETECTOR_CLASSES: list[type[adcore.AreaDetector]] = [
    adsimdetector.SimDetector,
    advimba.VimbaDetector,
    adandor.Andor2Detector,
    adaravis.AravisDetector,
    adkinetix.KinetixDetector,
    adpilatus.PilatusDetector,
    adcore.ContAcqAreaDetector,
]

WRITER_CLASSES: list[type[adcore.ADWriter]] = [
    adcore.ADHDFWriter,
    adcore.ADTIFFWriter,
    adcore.ADJPEGWriter,
]

pytestmark = pytest.mark.parametrize(
    "detector_cls, writer_cls",
    list(itertools.product(DETECTOR_CLASSES, WRITER_CLASSES)),
)


async def test_hints_from_hdf_writer(
    ad_standard_det_factory,
    detector_cls: type[adcore.AreaDetector],
    writer_cls: type[adcore.ADWriter],
):
    test_det = ad_standard_det_factory(detector_cls, writer_cls=writer_cls)
    assert test_det.hints == {"fields": [test_det.name]}


async def test_can_read(
    ad_standard_det_factory,
    detector_cls: type[adcore.AreaDetector],
    writer_cls: type[adcore.ADWriter],
):
    # Standard detector can be used as Readable
    test_det = ad_standard_det_factory(detector_cls, writer_cls=writer_cls)
    assert (await test_det.read()) == {}


async def test_describes_writer_dataset(
    ad_standard_det_factory,
    detector_cls: type[adcore.AreaDetector],
    writer_cls: type[adcore.ADWriter],
    one_shot_trigger_info: TriggerInfo,
):
    test_det = ad_standard_det_factory(detector_cls, writer_cls=writer_cls)
    assert await test_det.describe() == {}
    await test_det.stage()
    await test_det.prepare(one_shot_trigger_info)
    assert await test_det.describe() == {
        f"{test_det.name}": {
            "source": (
                f"mock+ca://{detector_cls.__name__[: -len('Detector')].upper()}"
                f"1:{writer_cls.default_suffix}FullFileName_RBV"
            ),
            "shape": [one_shot_trigger_info.exposures_per_event, 10, 10],
            "dtype": "array",
            "dtype_numpy": "<u2",
            "external": "STREAM:",
        }
    }


async def test_can_collect(
    ad_standard_det_factory,
    detector_cls: type[adcore.AreaDetector],
    writer_cls: type[adcore.ADWriter],
    static_path_provider: StaticPathProvider,
    one_shot_trigger_info: TriggerInfo,
):
    path_info = static_path_provider()

    test_det = ad_standard_det_factory(detector_cls, writer_cls=writer_cls)

    await test_det.stage()
    await test_det.prepare(one_shot_trigger_info)
    docs = [(name, doc) async for name, doc in test_det.collect_asset_docs(1)]
    assert len(docs) == 2
    assert docs[0][0] == "stream_resource"
    stream_resource = docs[0][1]
    sr_uid = stream_resource["uid"]
    assert stream_resource["data_key"] == test_det.name

    # With HDF writer, the URI points directly to the file. For other writers, since a
    # dataset is many files, point to the directory instead.
    expected_uri = (
        "file://localhost/"
        + path_info.directory_path.absolute().as_posix().lstrip("/")
        + "/"
    )
    if writer_cls == adcore.ADHDFWriter:
        expected_uri += path_info.filename + ".h5"

    assert stream_resource["uri"] == expected_uri

    # Construct expected stream resource parameters based on writer
    expected_sres_params: dict[str, Any] = {
        "chunk_shape": (1, 10, 10),
    }
    if writer_cls == adcore.ADHDFWriter:
        expected_sres_params["dataset"] = "/entry/data/data"
    else:
        expected_sres_params["template"] = (
            "ophyd_async_tests_{:06d}" + test_det._writer._file_extension
        )

    assert stream_resource["parameters"] == expected_sres_params
    assert docs[1][0] == "stream_datum"
    stream_datum = docs[1][1]
    assert stream_datum["stream_resource"] == sr_uid
    assert stream_datum["seq_nums"] == {"start": 0, "stop": 0}
    assert stream_datum["indices"] == {"start": 0, "stop": 1}


async def test_can_specify_different_uri_and_path(
    ad_standard_det_factory,
    detector_cls: type[adcore.AreaDetector],
    writer_cls: type[adcore.ADWriter],
    one_shot_trigger_info: TriggerInfo,
    tmp_path: Path,
    static_path_provider_factory,
    static_filename_provider,
):
    # Create a static path provider that will return a specific directory
    expected_uri = f"file://localhost/{tmp_path.absolute().as_posix()}/different/"
    path_provider = static_path_provider_factory(
        static_filename_provider, directory_uri=expected_uri
    )
    test_det = ad_standard_det_factory(
        detector_cls, writer_cls=writer_cls, path_provider=path_provider
    )

    path_info = test_det._writer._path_provider(device_name=test_det.name)

    await test_det.stage()
    await test_det.prepare(one_shot_trigger_info)
    docs = [(name, doc) async for name, doc in test_det.collect_asset_docs(1)]

    # Make sure we set the write path to the directory_path attr from our path info
    assert await test_det.fileio.file_path.get_value() == str(tmp_path) + os.sep

    # Then, check to make sure our resource doc uses the overridden URI
    assert len(docs) == 2
    assert docs[0][0] == "stream_resource"
    stream_resource = docs[0][1]

    # With HDF writer, the URI points directly to the file. For other writers, since a
    # dataset is many files, point to the directory instead.
    if writer_cls == adcore.ADHDFWriter:
        expected_uri += path_info.filename + ".h5"

    assert stream_resource["uri"] == expected_uri


@pytest.mark.parametrize("write_path_format", [PurePosixPath, PureWindowsPath])
async def test_can_override_uri_with_different_path_semantics(
    write_path_format,
    ad_standard_det_factory,
    detector_cls: type[adcore.AreaDetector],
    writer_cls: type[adcore.ADWriter],
    one_shot_trigger_info: TriggerInfo,
    static_filename_provider,
):
    windows_path = PureWindowsPath(
        "C:\\Users\\test\\AppData\\Local\\Temp\\ophyd_async_tests"
    )
    posix_path = PurePosixPath("/tmp/ophyd_async_tests")

    if write_path_format is PureWindowsPath:
        write_path = windows_path
        expected_uri = f"file://localhost/{posix_path.as_posix().lstrip('/')}/"
    else:
        write_path = posix_path
        expected_uri = f"file://localhost/{windows_path.as_posix().lstrip('/')}/"

    path_provider = StaticPathProvider(
        static_filename_provider, write_path, directory_uri=expected_uri
    )

    test_det = ad_standard_det_factory(
        detector_cls,
        writer_cls=writer_cls,
        path_provider=path_provider,
        assume_file_path_exists=True,
    )

    path_info = test_det._writer._path_provider(device_name=test_det.name)
    await test_det.stage()
    await test_det.prepare(one_shot_trigger_info)
    docs = [(name, doc) async for name, doc in test_det.collect_asset_docs(1)]
    if write_path_format is PureWindowsPath:
        assert await test_det.fileio.file_path.get_value() == f"{windows_path}\\"
    else:
        assert await test_det.fileio.file_path.get_value() == f"{posix_path}/"
    assert len(docs) == 2
    assert docs[0][0] == "stream_resource"
    stream_resource = docs[0][1]

    # With HDF writer, the URI points directly to the file. For other writers, since a
    # dataset is many files, point to the directory instead.
    if writer_cls == adcore.ADHDFWriter:
        expected_uri += path_info.filename + ".h5"

    assert stream_resource["uri"] == expected_uri


async def test_can_decribe_collect(
    ad_standard_det_factory,
    detector_cls: type[adcore.AreaDetector],
    writer_cls: type[adcore.ADWriter],
    one_shot_trigger_info: TriggerInfo,
):
    test_det = ad_standard_det_factory(detector_cls, writer_cls=writer_cls)

    assert (await test_det.describe_collect()) == {}
    await test_det.stage()
    await test_det.prepare(one_shot_trigger_info)
    assert (await test_det.describe_collect()) == {
        f"{test_det.name}": {
            "source": (
                f"mock+ca://{detector_cls.__name__[: -len('Detector')].upper()}"
                f"1:{writer_cls.default_suffix}FullFileName_RBV"
            ),
            "shape": [one_shot_trigger_info.exposures_per_event, 10, 10],
            "dtype": "array",
            "dtype_numpy": "<u2",
            "external": "STREAM:",
        }
    }



import asyncio
import os
import pprint
import subprocess
import sys
import time
from collections.abc import Callable
from pathlib import Path
from typing import Any

import pytest
from bluesky.run_engine import RunEngine, TransitionError
from bluesky.utils import new_uid
from pytest import FixtureRequest

from ophyd_async.core import (
    DetectorTrigger,
    FilenameProvider,
    StaticFilenameProvider,
    StaticPathProvider,
    TriggerInfo,
    init_devices,
)
from ophyd_async.epics import adsimdetector

PANDA_RECORD = str(
    Path(__file__).parent / "unit_tests" / "fastcs" / "panda" / "db" / "panda.db"
)
INCOMPLETE_BLOCK_RECORD = str(
    Path(__file__).parent
    / "unit_tests"
    / "fastcs"
    / "panda"
    / "db"
    / "incomplete_block_panda.db"
)
INCOMPLETE_RECORD = str(
    Path(__file__).parent
    / "unit_tests"
    / "fastcs"
    / "panda"
    / "db"
    / "incomplete_panda.db"
)
EXTRA_BLOCKS_RECORD = str(
    Path(__file__).parent
    / "unit_tests"
    / "fastcs"
    / "panda"
    / "db"
    / "extra_blocks_panda.db"
)


# Autouse fixture that will set all EPICS networking env vars to use lo interface
# to avoid false failures caused by things like firewalls blocking EPICS traffic.
@pytest.fixture(scope="session", autouse=True)
def configure_epics_environment():
    os.environ["EPICS_CAS_INTF_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_CAS_BEACON_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_CA_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_CAS_AUTO_ADDR_LIST"] = "NO"
    os.environ["EPICS_CA_AUTO_BEACON_ADDR_LIST"] = "NO"

    os.environ["EPICS_PVAS_INTF_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_PVAS_BEACON_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_PVA_ADDR_LIST"] = "127.0.0.1"
    os.environ["EPICS_PVAS_AUTO_BEACON_ADDR_LIST"] = "NO"
    os.environ["EPICS_PVA_AUTO_ADDR_LIST"] = "NO"


_ALLOWED_PYTEST_TASKS = {"async_finalizer", "async_setup", "async_teardown"}


def _error_and_kill_pending_tasks(
    loop: asyncio.AbstractEventLoop, test_name: str, test_passed: bool
) -> set[asyncio.Task]:
    """Cancels pending tasks in the event loop for a test. Raises an exception if
    the test hasn't already.

    Args:
        loop: The event loop to check for pending tasks.
        test_name: The name of the test.
        test_passed: Indicates whether the test passed.

    Returns:
        set[asyncio.Task]: The set of unfinished tasks that were cancelled.

    Raises:
        RuntimeError: If there are unfinished tasks and the test didn't fail.
    """
    unfinished_tasks = {
        task
        for task in asyncio.all_tasks(loop)
        if (coro := task.get_coro()) is not None
        and hasattr(coro, "__name__")
        and coro.__name__ not in _ALLOWED_PYTEST_TASKS
        and not task.done()
    }
    for task in unfinished_tasks:
        task.cancel()

    # We only raise an exception here if the test didn't fail anyway.
    # If it did then it makes sense that there's some tasks we need to cancel,
    # but an exception will already have been raised.
    if unfinished_tasks and test_passed:
        raise RuntimeError(
            f"Not all tasks closed during test {test_name}:\n"
            f"{pprint.pformat(unfinished_tasks, width=88)}"
        )

    return unfinished_tasks


@pytest.fixture(autouse=True, scope="function")
async def fail_test_on_unclosed_tasks(request: FixtureRequest):
    """Used on every test to ensure failure if there are pending tasks
    by the end of the test.
    """
    try:
        fail_count = request.session.testsfailed
        loop = asyncio.get_running_loop()

        loop.set_debug(True)

        request.addfinalizer(
            lambda: _error_and_kill_pending_tasks(
                loop, request.node.name, request.session.testsfailed == fail_count
            )
        )
    # Once https://github.com/bluesky/ophyd-async/issues/683
    # is finished we can remove this try, except.
    except RuntimeError as error:
        if str(error) != "no running event loop":
            raise error


@pytest.fixture(scope="function")
def RE(request: FixtureRequest):
    loop = asyncio.new_event_loop()
    loop.set_debug(True)
    RE = RunEngine({}, call_returns_result=True, loop=loop)
    fail_count = request.session.testsfailed

    def clean_event_loop():
        if RE.state not in ("idle", "panicked"):
            try:
                RE.halt()
            except TransitionError:
                pass

        loop.call_soon_threadsafe(loop.stop)
        RE._th.join()

        try:
            _error_and_kill_pending_tasks(
                loop, request.node.name, request.session.testsfailed == fail_count
            )
        finally:
            loop.close()

    request.addfinalizer(clean_event_loop)
    return RE


@pytest.fixture(scope="module", params=["pva"])
def panda_pva():
    processes = [
        subprocess.Popen(
            [
                sys.executable,
                "-m",
                "epicscorelibs.ioc",
                "-m",
                macros,
                "-d",
                PANDA_RECORD,
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        for macros in [
            "INCLUDE_EXTRA_BLOCK=,INCLUDE_EXTRA_SIGNAL=",
            "EXCLUDE_WIDTH=#,IOC_NAME=PANDAQSRVIB",
            "EXCLUDE_PCAP=#,IOC_NAME=PANDAQSRVI",
        ]
    ]
    time.sleep(2)

    for p in processes:
        assert not p.poll(), p.stdout.read()

    yield processes

    for p in processes:
        p.terminate()
        p.communicate()


@pytest.fixture
async def normal_coroutine() -> Callable[[], Any]:
    async def inner_coroutine():
        await asyncio.sleep(0.01)

    return inner_coroutine


@pytest.fixture
async def failing_coroutine() -> Callable[[], Any]:
    async def inner_coroutine():
        await asyncio.sleep(0.01)
        raise ValueError()

    return inner_coroutine


@pytest.fixture
def static_filename_provider():
    return StaticFilenameProvider("ophyd_async_tests")


@pytest.fixture
def static_path_provider_factory(tmp_path: Path):
    def create_static_dir_provider_given_fp(
        fp: FilenameProvider, directory_uri: str | None = None
    ):
        return StaticPathProvider(fp, tmp_path, directory_uri=directory_uri)

    return create_static_dir_provider_given_fp


@pytest.fixture
def static_path_provider(
    static_path_provider_factory: Callable,
    static_filename_provider: FilenameProvider,
):
    return static_path_provider_factory(static_filename_provider)


@pytest.fixture
def one_shot_trigger_info(request: FixtureRequest) -> TriggerInfo:
    # If the fixture is called with a parameter, use it as the exposures_per_event
    # otherwise use 1
    param = getattr(request, "param", 1)
    return TriggerInfo(
        number_of_events=1,
        trigger=DetectorTrigger.INTERNAL,
        livetime=None,
        exposures_per_event=param if isinstance(param, int) else 1,
    )


@pytest.fixture
async def sim_detector(request: FixtureRequest):
    """Fixture that creates a simulated detector.

    Args:
        prefix (str): The PV prefix for the detector
        name (str): Name for the detector instance
        tmp_path (Path): Temporary directory for file writing
    """
    prefix = (
        request.param[0] if isinstance(request.param, list | tuple) else request.param
    )
    name = request.param[1] if isinstance(request.param, list | tuple) else "test"
    tmp_path = request.getfixturevalue("tmp_path")

    fp = StaticFilenameProvider(f"test-{new_uid()}")
    dp = StaticPathProvider(fp, tmp_path)

    async with init_devices(mock=True):
        det = adsimdetector.SimDetector(prefix, dp, name=name)

    det._config_sigs = [det.driver.acquire_time, det.driver.acquire]

    return det



from collections.abc import Sequence
from typing import Any

import numpy as np

from ophyd_async.core import (
    Array1D,
    Device,
    DeviceVector,
    DTypeScalar_co,
    SignalRW,
    StandardReadable,
    StrictEnum,
    Table,
    soft_signal_r_and_setter,
    soft_signal_rw,
)
from ophyd_async.core import StandardReadableFormat as Format


class ExampleEnum(StrictEnum):
    """Example of a strict Enum datatype."""

    A = "Aaa"
    B = "Bbb"
    C = "Ccc"


class ExampleTable(Table):
    a_bool: Array1D[np.bool_]
    a_int: Array1D[np.int32]
    a_float: Array1D[np.float64]
    a_str: Sequence[str]
    a_enum: Sequence[ExampleEnum]


def int_array_value(dtype: type[DTypeScalar_co]):
    iinfo = np.iinfo(dtype)  # type: ignore
    return np.array([iinfo.min, iinfo.max, 0, 1, 2, 3, 4], dtype=dtype)


def int_array_signal(
    dtype: type[DTypeScalar_co], name: str = ""
) -> SignalRW[Array1D[DTypeScalar_co]]:
    value = int_array_value(dtype)
    return soft_signal_rw(Array1D[dtype], value, name)


def float_array_value(dtype: type[DTypeScalar_co]):
    finfo = np.finfo(dtype)  # type: ignore
    return np.array(
        [
            finfo.min,
            finfo.max,
            finfo.smallest_normal,
            finfo.smallest_subnormal,
            0,
            1.234,
            2.34e5,
            3.45e-6,
        ],
        dtype=dtype,
    )


def float_array_signal(
    dtype: type[DTypeScalar_co], name: str = ""
) -> SignalRW[Array1D[DTypeScalar_co]]:
    value = float_array_value(dtype)
    return soft_signal_rw(Array1D[dtype], value, name)


class OneOfEverythingDevice(StandardReadable):
    """A device with one of every datatype allowed on signals."""

    # make a detector to test assert_configuration
    def __init__(self, name=""):
        # add all signals to configuration
        with self.add_children_as_readables(Format.CONFIG_SIGNAL):
            self.a_int = soft_signal_rw(int, 1)
            self.a_float = soft_signal_rw(float, 1.234)
            self.a_str = soft_signal_rw(str, "test_string")
            self.a_bool = soft_signal_rw(bool, True)
            self.a_enum = soft_signal_rw(ExampleEnum, ExampleEnum.B)
            self.boola = soft_signal_rw(
                Array1D[np.bool_], np.array([False, False, True])
            )
            self.int8a = int_array_signal(np.int8)
            self.uint8a = int_array_signal(np.uint8)
            self.int16a = int_array_signal(np.int16)
            self.uint16a = int_array_signal(np.uint16)
            self.int32a = int_array_signal(np.int32)
            self.uint32a = int_array_signal(np.uint32)
            self.int64a = int_array_signal(np.int64)
            self.uint64a = int_array_signal(np.uint64)
            self.float32a = float_array_signal(np.float32)
            self.float64a = float_array_signal(np.float64)
            self.stra = soft_signal_rw(
                Sequence[str],
                ["one", "two", "three"],
            )
            self.enuma = soft_signal_rw(
                Sequence[ExampleEnum],
                [ExampleEnum.A, ExampleEnum.C],
            )
            self.table = soft_signal_rw(
                ExampleTable,
                ExampleTable(
                    a_bool=np.array([False, False, True, True], np.bool_),
                    a_int=np.array([1, 8, -9, 32], np.int32),
                    a_float=np.array([1.8, 8.2, -6, 32.9887], np.float64),
                    a_str=["Hello", "World", "Foo", "Bar"],
                    a_enum=[ExampleEnum.A, ExampleEnum.B, ExampleEnum.A, ExampleEnum.C],
                ),
            )
            self.ndarray = soft_signal_rw(np.ndarray, np.array(([1, 2, 3], [4, 5, 6])))
        super().__init__(name)

    async def get_signal_values(self):
        return await _get_signal_values(self)


async def _get_signal_values(child: Device) -> dict[SignalRW, Any]:
    if isinstance(child, SignalRW):
        return {child: await child.get_value()}
    ret = {}
    for _, c in child.children():
        ret.update(await _get_signal_values(c))
    return ret


class ParentOfEverythingDevice(Device):
    """Device containing subdevices with one of every datatype allowed on signals."""

    def __init__(self, name=""):
        self.child = OneOfEverythingDevice()
        self.vector = DeviceVector(
            {1: OneOfEverythingDevice(), 3: OneOfEverythingDevice()}
        )
        self.sig_rw = soft_signal_rw(str, "Top level SignalRW")
        self.sig_r, _ = soft_signal_r_and_setter(str, "Top level SignalR")
        self._sig_rw = soft_signal_rw(str, "Top level private SignalRW")
        super().__init__(name=name)

    async def get_signal_values(self):
        return await _get_signal_values(self)



"""Used for tutorial `Implementing Devices`."""

# Import bluesky and ophyd
import bluesky.plan_stubs as bps  # noqa: F401
import bluesky.plans as bp  # noqa: F401
from bluesky.callbacks.best_effort import BestEffortCallback
from bluesky.run_engine import RunEngine, autoawait_in_bluesky_event_loop

from ophyd_async.core import init_devices
from ophyd_async.epics import demo, testing

# Create a run engine and make ipython use it for `await` commands
RE = RunEngine(call_returns_result=True)
autoawait_in_bluesky_event_loop()

# Add a callback for plotting
bec = BestEffortCallback()
RE.subscribe(bec)

# Start IOC with demo pvs in subprocess
prefix = testing.generate_random_pv_prefix()
ioc = demo.start_ioc_subprocess(prefix, num_channels=3)

# All Devices created within this block will be
# connected and named at the end of the with block
with init_devices():
    # Create a sample stage with X and Y motors
    stage = demo.DemoStage(f"{prefix}STAGE:")
    # Create a multi channel counter with the same number
    # of counters as the IOC
    pdet = demo.DemoPointDetector(f"{prefix}DET:", num_channels=3)



from __future__ import annotations

import asyncio
from collections.abc import Callable, Mapping
from typing import Any

import bluesky.plan_stubs as bps
import numpy as np
from bluesky.utils import MsgGenerator, plan

from ophyd_async.core import (
    Device,
    Settings,
    SettingsProvider,
    SignalRW,
    Table,
    walk_config_signals,
    walk_rw_signals,
)

from ._utils import T
from ._wait_for_awaitable import wait_for_awaitable


@plan
def _get_values_of_signals(
    signals: Mapping[T, SignalRW],
) -> MsgGenerator[dict[T, Any]]:
    coros = [sig.get_value() for sig in signals.values()]
    values = yield from wait_for_awaitable(asyncio.gather(*coros))
    named_values = dict(zip(signals, values, strict=True))
    return named_values


@plan
def get_current_settings(
    device: Device, only_config: bool = False
) -> MsgGenerator[Settings]:
    """Get current settings on `Device`.

    If `only_config` is True, get current configuration settings on `Configurable`.
    """
    if only_config:
        signals = yield from wait_for_awaitable(walk_config_signals(device))
    else:
        signals = walk_rw_signals(device)
    named_values = yield from _get_values_of_signals(signals)
    signal_values = {signals[name]: value for name, value in named_values.items()}
    return Settings(device, signal_values)


@plan
def store_settings(
    provider: SettingsProvider, name: str, device: Device, only_config: bool = False
) -> MsgGenerator[None]:
    """Walk a Device for SignalRWs and store their values.

    If `only_config` is True, store only configuration settings on `Configurable`.

    :param provider: The provider to store the settings with.
    :param name: The name to store the settings under.
    :param device: The Device to walk for SignalRWs.
    :param only_config: If True, store only configuration settings.
    """
    if only_config:
        signals = yield from wait_for_awaitable(walk_config_signals(device))
    else:
        signals = walk_rw_signals(device)
    named_values = yield from _get_values_of_signals(signals)
    yield from wait_for_awaitable(provider.store(name, named_values))


@plan
def retrieve_settings(
    provider: SettingsProvider, name: str, device: Device, only_config: bool = False
) -> MsgGenerator[Settings]:
    """Retrieve named Settings for a Device from a provider.

    If `only_config` is True, retrieve only configuration settings on `Configurable`.

    :param provider: The provider to retrieve the settings from.
    :param name: The name of the settings to retrieve.
    :param device: The Device to retrieve the settings for.
    :param only_config: If True, retrieve only configuration settings.
    """
    named_values = yield from wait_for_awaitable(provider.retrieve(name))
    if only_config:
        signals = yield from wait_for_awaitable(walk_config_signals(device))
    else:
        signals = walk_rw_signals(device)
    unknown_names = set(named_values) - set(signals)
    if unknown_names:
        raise NameError(f"Unknown signal names {sorted(unknown_names)}")
    signal_values = {signals[name]: value for name, value in named_values.items()}
    return Settings(device, signal_values)


@plan
def apply_settings(settings: Settings) -> MsgGenerator[None]:
    """Set every SignalRW to the given value in Settings. If value is None ignore it."""
    signal_values = {
        signal: value for signal, value in settings.items() if value is not None
    }
    if signal_values:
        for signal, value in signal_values.items():
            yield from bps.abs_set(signal, value, group="apply_settings")
        yield from bps.wait("apply_settings")


@plan
def apply_settings_if_different(
    settings: Settings,
    apply_plan: Callable[[Settings], MsgGenerator[None]],
    current_settings: Settings | None = None,
) -> MsgGenerator[None]:
    """Set every SignalRW in settings, only if it is different to the current value.

    :param apply_plan:
        A device specific plan which takes the Settings to apply and applies them to
        the Device. Used to add device specific ordering to setting the signals.
    :param current_settings:
        If given, should be a superset of settings containing the current value of
        the Settings in the Device. If not given it will be created by reading just
        the signals given in settings.
    """
    if current_settings is None:
        # If we aren't give the current settings, then get the
        # values of just the signals we were asked to change.
        # This allows us to use this plan with Settings for a subset
        # of signals in the Device without retrieving them all
        signal_values = yield from _get_values_of_signals(
            {sig: sig for sig in settings}
        )
        current_settings = Settings(settings.device, signal_values)

    def _is_different(current, required) -> bool:
        if isinstance(current, Table):
            current = current.model_dump()
            if isinstance(required, Table):
                required = required.model_dump()
            return current.keys() != required.keys() or any(
                _is_different(current[k], required[k]) for k in current
            )
        elif isinstance(current, np.ndarray):
            return not np.array_equal(current, required)
        else:
            return current != required

    settings_to_change, _ = settings.partition(
        lambda sig: _is_different(current_settings[sig], settings[sig])
    )
    yield from apply_plan(settings_to_change)



import uuid
from abc import abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from datetime import date
from pathlib import PurePath, PureWindowsPath
from typing import Protocol
from urllib.parse import urlunparse


@dataclass
class PathInfo:
    """Information about where and how to write a file.

    :param directory_path: Directory into which files should be written
    :param filename: Base filename to use generated by FilenameProvider, w/o extension
    :param create_dir_depth: Optional depth of directories to create if they do not
                             exist
    :param directory_uri: Optional URI to use for reading back resources. If not set,
                          it will be generated from the directory path.
    """

    directory_path: PurePath
    filename: str
    create_dir_depth: int = 0
    directory_uri: str | None = None

    def __post_init__(self):
        if not self.directory_path.is_absolute():
            raise ValueError(
                f"directory_path must be an absolute path, got {self.directory_path}"
            )

        # If directory uri is not set, set it using the directory path.
        if self.directory_uri is None:
            self.directory_uri = urlunparse(
                (
                    "file",
                    "localhost",
                    f"{self.directory_path.as_posix()}/",
                    "",
                    "",
                    None,
                )
            )
        elif not self.directory_uri.endswith("/"):
            # Ensure the directory URI ends with a slash.
            self.directory_uri += "/"


class FilenameProvider(Protocol):
    """Base class for callable classes providing filenames."""

    @abstractmethod
    def __call__(self, device_name: str | None = None) -> str:
        """Get a filename to use for output data, w/o extension."""


class PathProvider(Protocol):
    """Abstract class that tells a detector where to write its data."""

    @abstractmethod
    def __call__(self, device_name: str | None = None) -> PathInfo:
        """Get the current directory to write files into."""


class StaticFilenameProvider(FilenameProvider):
    """Provides a constant filename on every call."""

    def __init__(self, filename: str):
        self._static_filename = filename

    def __call__(self, device_name: str | None = None) -> str:
        return self._static_filename


class UUIDFilenameProvider(FilenameProvider):
    """Files will have a UUID as a filename."""

    def __init__(
        self,
        uuid_call_func: Callable = uuid.uuid4,
        uuid_call_args: list | None = None,
    ):
        self._uuid_call_func = uuid_call_func
        self._uuid_call_args = uuid_call_args or []

    def __call__(self, device_name: str | None = None) -> str:
        if (
            self._uuid_call_func in [uuid.uuid3, uuid.uuid5]
            and len(self._uuid_call_args) < 2
        ):
            raise ValueError(
                f"To use {self._uuid_call_func} to generate UUID filenames,"
                " UUID namespace and name must be passed as args!"
            )

        uuid_str = self._uuid_call_func(*self._uuid_call_args)
        return f"{uuid_str}"


class AutoIncrementFilenameProvider(FilenameProvider):
    """Provides a new numerically incremented filename on each call."""

    def __init__(
        self,
        base_filename: str = "",
        max_digits: int = 5,
        starting_value: int = 0,
        increment: int = 1,
        inc_delimeter: str = "_",
    ):
        self._base_filename = base_filename
        self._max_digits = max_digits
        self._current_value = starting_value
        self._increment = increment
        self._inc_delimeter = inc_delimeter

    def __call__(self, device_name: str | None = None) -> str:
        if len(str(self._current_value)) > self._max_digits:
            raise ValueError(
                f"Auto incrementing filename counter \
                  exceeded maximum of {self._max_digits} digits!"
            )

        padded_counter = f"{self._current_value:0{self._max_digits}}"

        filename = f"{self._base_filename}{self._inc_delimeter}{padded_counter}"

        self._current_value += self._increment
        return filename


class StaticPathProvider(PathProvider):
    """All files will be within a static directory."""

    def __init__(
        self,
        filename_provider: FilenameProvider,
        directory_path: PurePath,
        directory_uri: str | None = None,
        create_dir_depth: int = 0,
    ) -> None:
        self._filename_provider = filename_provider
        self._directory_path = directory_path
        self._directory_uri = directory_uri
        self._create_dir_depth = create_dir_depth

    def __call__(self, device_name: str | None = None) -> PathInfo:
        filename = self._filename_provider(device_name)

        return PathInfo(
            directory_path=self._directory_path,
            directory_uri=self._directory_uri,
            filename=filename,
            create_dir_depth=self._create_dir_depth,
        )


class AutoIncrementingPathProvider(PathProvider):
    """Provides a new numerically incremented path on each call."""

    def __init__(
        self,
        filename_provider: FilenameProvider,
        base_directory_path: PurePath,
        base_directory_uri: str | None = None,
        create_dir_depth: int = 0,
        max_digits: int = 5,
        starting_value: int = 0,
        num_calls_per_inc: int = 1,
        increment: int = 1,
        inc_delimeter: str = "_",
        base_name: str | None = None,
    ) -> None:
        self._filename_provider = filename_provider
        self._base_directory_path = base_directory_path
        self._base_directory_uri = base_directory_uri
        if (
            self._base_directory_uri is not None
            and not self._base_directory_uri.endswith("/")
        ):
            self._base_directory_uri += "/"
        self._create_dir_depth = create_dir_depth
        self._base_name = base_name
        self._starting_value = starting_value
        self._current_value = starting_value
        self._num_calls_per_inc = num_calls_per_inc
        self._inc_counter = 0
        self._max_digits = max_digits
        self._increment = increment
        self._inc_delimeter = inc_delimeter

    def __call__(self, device_name: str | None = None) -> PathInfo:
        filename = self._filename_provider(device_name)

        padded_counter = f"{self._current_value:0{self._max_digits}}"

        auto_inc_dir_name = str(padded_counter)
        if self._base_name is not None:
            auto_inc_dir_name = (
                f"{self._base_name}{self._inc_delimeter}{padded_counter}"
            )
        elif device_name is not None:
            auto_inc_dir_name = f"{device_name}{self._inc_delimeter}{padded_counter}"

        self._inc_counter += 1
        if self._inc_counter == self._num_calls_per_inc:
            self._inc_counter = 0
            self._current_value += self._increment

        directory_uri = None
        if self._base_directory_uri is not None:
            directory_uri = f"{self._base_directory_uri}{auto_inc_dir_name}"

        return PathInfo(
            directory_path=self._base_directory_path / auto_inc_dir_name,
            directory_uri=directory_uri,
            filename=filename,
            create_dir_depth=self._create_dir_depth,
        )


class YMDPathProvider(PathProvider):
    """Provides a path with the date included in the directory name."""

    def __init__(
        self,
        filename_provider: FilenameProvider,
        base_directory_path: PurePath,
        base_directory_uri: str | None = None,
        create_dir_depth: int = -3,  # Default to -3 to create YMD dirs
        device_name_as_base_dir: bool = False,
    ) -> None:
        self._filename_provider = filename_provider
        self._base_directory_path = base_directory_path
        self._base_directory_uri = base_directory_uri
        if (
            self._base_directory_uri is not None
            and not self._base_directory_uri.endswith("/")
        ):
            self._base_directory_uri += "/"
        self._create_dir_depth = create_dir_depth
        self._device_name_as_base_dir = device_name_as_base_dir

    def __call__(self, device_name: str | None = None) -> PathInfo:
        path_type = type(self._base_directory_path)
        if path_type == PureWindowsPath:
            sep = "\\"
        else:
            sep = "/"

        current_date = date.today().strftime(f"%Y{sep}%m{sep}%d")
        if device_name is None:
            ymd_dir_path = current_date
        elif self._device_name_as_base_dir:
            ymd_dir_path = path_type(
                current_date,
                device_name,
            )
        else:
            ymd_dir_path = path_type(
                device_name,
                current_date,
            )

        filename = self._filename_provider(device_name)

        directory_uri = None
        if self._base_directory_uri is not None:
            directory_uri = f"{self._base_directory_uri}{ymd_dir_path}"

        return PathInfo(
            directory_path=self._base_directory_path / ymd_dir_path,
            directory_uri=directory_uri,
            filename=filename,
            create_dir_depth=self._create_dir_depth,
        )


class DatasetDescriber(Protocol):
    """For describing datasets in file writing."""

    @abstractmethod
    async def np_datatype(self) -> str:
        """Return the numpy datatype for this dataset."""

    @abstractmethod
    async def shape(self) -> tuple[int | None, ...]:
        """Get the shape of the data collection."""



from ophyd_async.core import DatasetDescriber, PathProvider

from ._core_io import NDFilePluginIO, NDPluginBaseIO
from ._core_writer import ADWriter


class ADTIFFWriter(ADWriter[NDFilePluginIO]):
    default_suffix: str = "TIFF1:"

    def __init__(
        self,
        fileio: NDFilePluginIO,
        path_provider: PathProvider,
        dataset_describer: DatasetDescriber,
        plugins: dict[str, NDPluginBaseIO] | None = None,
    ) -> None:
        super().__init__(
            fileio,
            path_provider,
            dataset_describer,
            plugins=plugins,
            file_extension=".tiff",
            mimetype="multipart/related;type=image/tiff",
        )



import asyncio
from contextlib import suppress

from ophyd_async.core import DetectorController, DetectorTrigger, TriggerInfo

from ._pattern_generator import PatternGenerator


class BlobDetectorController(DetectorController):
    def __init__(self, pattern_generator: PatternGenerator):
        self.pattern_generator = pattern_generator
        self.trigger_info: TriggerInfo | None = None
        self.task: asyncio.Task | None = None

    def get_deadtime(self, exposure):
        return 0.001

    async def prepare(self, trigger_info: TriggerInfo):
        # This is a simulation, so only support intenal triggering
        if trigger_info.trigger != DetectorTrigger.INTERNAL:
            raise RuntimeError(f"{trigger_info.trigger} not supported by {self}")
        # Just hold onto the trigger info until we need it
        self.trigger_info = trigger_info

    async def arm(self):
        if self.trigger_info is None:
            raise RuntimeError(f"prepare() not called on {self}")
        livetime = self.trigger_info.livetime or 0.1
        # Start a background process off writing the images to file
        coro = self.pattern_generator.write_images_to_file(
            exposure=livetime,
            period=livetime + self.trigger_info.deadtime,
            number_of_frames=self.trigger_info.total_number_of_exposures,
        )
        self.task = asyncio.create_task(coro)

    async def wait_for_idle(self):
        # Wait for the background task to complete
        if self.task:
            await self.task

    async def disarm(self):
        # Stop the background task and wait for it to finish
        if self.task:
            self.task.cancel()
            with suppress(asyncio.CancelledError):
                await self.task
            self.task = None



from ._example_ioc import (
    CA_PVA_RECORDS,
    PVA_RECORDS,
    EpicsTestCaDevice,
    EpicsTestEnum,
    EpicsTestIocAndDevices,
    EpicsTestPvaDevice,
    EpicsTestSubsetEnum,
    EpicsTestTable,
)
from ._utils import TestingIOC, generate_random_pv_prefix

__all__ = [
    "CA_PVA_RECORDS",
    "PVA_RECORDS",
    "EpicsTestCaDevice",
    "EpicsTestEnum",
    "EpicsTestSubsetEnum",
    "EpicsTestPvaDevice",
    "EpicsTestTable",
    "EpicsTestIocAndDevices",
    "TestingIOC",
    "generate_random_pv_prefix",
]



import asyncio
from pathlib import Path
from typing import Any
from unittest.mock import patch

import bluesky.plan_stubs as bps
import bluesky.plans as bp
import pytest
from bluesky.run_engine import RunEngine

from ophyd_async.core import (
    AsyncStatus,
    DetectorController,
    DetectorTrigger,
    FlyerController,
    StandardDetector,
    StandardFlyer,
    TriggerInfo,
    init_devices,
)
from ophyd_async.epics import adcore, adsimdetector
from ophyd_async.testing import set_mock_value


class DummyTriggerLogic(FlyerController[int]):
    def __init__(self): ...

    async def prepare(self, value: int):
        return value

    async def kickoff(self): ...

    async def complete(self): ...

    async def stop(self): ...


class DummyController(DetectorController):
    def __init__(self) -> None: ...
    async def prepare(self, trigger_info: TriggerInfo):
        await asyncio.sleep(0.01)

    async def arm(self):
        self._arm_status = AsyncStatus(asyncio.sleep(0.01))

    async def wait_for_idle(self):
        await self._arm_status

    async def disarm(self): ...

    def get_deadtime(self, exposure: float | None) -> float:
        return 0.002


@pytest.fixture
def controller(RE) -> adsimdetector.SimController:
    with init_devices(mock=True):
        drv = adcore.ADBaseIO("DRV")

    return adsimdetector.SimController(drv)


@pytest.fixture
def writer(RE, static_path_provider, tmp_path: Path) -> adcore.ADHDFWriter:
    with init_devices(mock=True):
        hdf = adcore.NDFileHDFIO("HDF")

    return adcore.ADHDFWriter(
        hdf,
        static_path_provider,
        adcore.ADBaseDatasetDescriber(hdf),
        {},
    )


@patch("ophyd_async.core._detector.DEFAULT_TIMEOUT", 0.01)
async def test_hdf_writer_fails_on_timeout_with_stepscan(
    RE: RunEngine,
    writer: adcore.ADHDFWriter,
    controller: adsimdetector.SimController,
):
    set_mock_value(writer.fileio.file_path_exists, True)
    detector: StandardDetector[Any, Any] = StandardDetector(
        controller, writer, name="detector"
    )

    with pytest.raises(Exception) as exc:
        RE(bp.count([detector]))

    assert isinstance(exc.value.__cause__, asyncio.TimeoutError)


@patch("ophyd_async.core._detector.DEFAULT_TIMEOUT", 0.01)
@pytest.mark.timeout(7)
def test_hdf_writer_fails_on_timeout_with_fly_scan(
    RE: RunEngine, writer: adcore.ADHDFWriter
):
    controller = DummyController()
    set_mock_value(writer.fileio.file_path_exists, True)

    detector: StandardDetector[Any, Any] = StandardDetector(controller, writer)
    trigger_logic = DummyTriggerLogic()

    flyer = StandardFlyer(trigger_logic, name="flyer")
    trigger_info = TriggerInfo(
        number_of_events=1,
        trigger=DetectorTrigger.CONSTANT_GATE,
        deadtime=2,
        livetime=2,
    )

    def flying_plan():
        """NOTE: the following is a workaround to ensure tests always pass.
        See https://github.com/bluesky/bluesky/issues/1630 for more details.
        """
        yield from bps.stage_all(detector, flyer)
        try:
            # Prepare the flyer first to get the trigger info for the detectors
            yield from bps.prepare(flyer, 1, wait=True)
            # prepare detector second.
            yield from bps.prepare(detector, trigger_info, wait=True)
            yield from bps.open_run()
            yield from bps.kickoff(flyer)
            yield from bps.kickoff(detector)
            yield from bps.complete(flyer, wait=True)
            yield from bps.complete(detector, wait=True)
            yield from bps.close_run()
        finally:
            yield from bps.unstage_all(detector, flyer)

    with pytest.raises(Exception) as exc:
        RE(flying_plan())

    assert isinstance(exc.value.__cause__, asyncio.TimeoutError)


async def test_ad_sim_controller_raise(controller: adsimdetector.SimController):
    with pytest.raises(
        TypeError,
        match=r"fly scanning .* is not supported for this device",
    ):
        await controller.prepare(
            TriggerInfo(number_of_events=1, trigger=DetectorTrigger.EDGE_TRIGGER)
        )



from collections.abc import Sequence

import numpy as np

from ophyd_async.core import Array1D, Device, DeviceVector, StandardReadable
from ophyd_async.epics import motor
from ophyd_async.epics.core import epics_signal_r, epics_signal_rw, epics_signal_x

CS_LETTERS = "ABCUVWXYZ"


class PmacTrajectoryIO(StandardReadable):
    """Device that moves a PMAC Motor record."""

    def __init__(self, prefix: str, name: str = "") -> None:
        self.time_array = epics_signal_rw(
            Array1D[np.float64], prefix + "ProfileTimeArray"
        )
        self.user_array = epics_signal_rw(Array1D[np.int32], prefix + "UserArray")
        # 1 indexed CS axes so we can index into them from the compound motor input link
        self.positions = DeviceVector(
            {
                i + 1: epics_signal_rw(
                    Array1D[np.float64], f"{prefix}{letter}:Positions"
                )
                for i, letter in enumerate(CS_LETTERS)
            }
        )
        self.use_axis = DeviceVector(
            {
                i + 1: epics_signal_rw(bool, f"{prefix}{letter}:UseAxis")
                for i, letter in enumerate(CS_LETTERS)
            }
        )
        self.velocities = DeviceVector(
            {
                i + 1: epics_signal_rw(
                    Array1D[np.float64], f"{prefix}{letter}:Velocities"
                )
                for i, letter in enumerate(CS_LETTERS)
            }
        )
        self.total_points = epics_signal_r(int, f"{prefix}TotalPoints_RBV")
        self.points_to_build = epics_signal_rw(int, prefix + "ProfilePointsToBuild")
        self.build_profile = epics_signal_x(prefix + "ProfileBuild")
        self.append_profile = epics_signal_x(prefix + "ProfileAppend")
        # This should be a SignalX, but because it is a Busy record, must
        # be a SignalRW to be waited on in PmacTrajectoryTriggerLogic.
        # TODO: Change record type to bo from busy (https://github.com/DiamondLightSource/pmac/issues/154)
        self.execute_profile = epics_signal_rw(bool, prefix + "ProfileExecute")
        self.abort_profile = epics_signal_x(prefix + "ProfileAbort")
        self.profile_cs_name = epics_signal_rw(str, prefix + "ProfileCsName")
        self.calculate_velocities = epics_signal_rw(bool, prefix + "ProfileCalcVel")

        super().__init__(name=name)


class PmacAxisAssignmentIO(Device):
    """A Device that (direct) moves a PMAC Coordinate System Motor.

    Note that this does not go through a motor record.
    """

    def __init__(self, prefix: str, name: str = "") -> None:
        self.cs_axis_letter = epics_signal_r(str, f"{prefix}CsAxis_RBV")
        self.cs_port = epics_signal_r(str, f"{prefix}CsPort_RBV")
        self.cs_number = epics_signal_r(int, f"{prefix}CsRaw_RBV")
        super().__init__(name=name)


class PmacCoordIO(Device):
    """A Device that represents a Pmac Coordinate System."""

    def __init__(self, prefix: str, name: str = "") -> None:
        self.defer_moves = epics_signal_rw(bool, f"{prefix}DeferMoves")
        self.cs_port = epics_signal_r(str, f"{prefix}Port")
        self.cs_axis_setpoint = DeviceVector(
            {
                i + 1: epics_signal_rw(float, f"{prefix}M{i + 1}:DirectDemand")
                for i in range(len(CS_LETTERS))
            }
        )
        super().__init__(name=name)


class PmacIO(Device):
    """Device that represents a pmac controller."""

    def __init__(
        self,
        prefix: str,
        raw_motors: Sequence[motor.Motor],
        coord_nums: Sequence[int],
        name: str = "",
    ) -> None:
        motor_prefixes = [motor.motor_egu.source.split(".")[0] for motor in raw_motors]

        self.assignment = DeviceVector(
            {
                i: PmacAxisAssignmentIO(motor_prefix + ":")
                for i, motor_prefix in enumerate(motor_prefixes)
            }
        )

        # Public Look up for motor to axis assignment DeviceVector index

        self.motor_assignment_index = {motor: i for i, motor in enumerate(raw_motors)}

        self.coord = DeviceVector(
            {coord: PmacCoordIO(prefix=f"{prefix}CS{coord}:") for coord in coord_nums}
        )
        # Trajectory PVs have the same prefix as the pmac device
        self.trajectory = PmacTrajectoryIO(prefix)

        super().__init__(name=name)
